const REST = require('./rest.js');
const WS = require('./ws.js');
const N = require('precise-number');
const R = require('ramda');
const EXCHANGE = require('../exchange.js');

class BINANCE extends EXCHANGE {
	constructor(options) {
		options = Object.assign({
			Name: 'Binance',
			Fees: {
				Maker: 0.002,
				Taker: 0.002
			},
			RateLimit: 10,
			MinTradeAmount: 0.1,
			DefaultDepthStep: 20
		}, options);
		options.domain = options.domain || 'api.binance.com';
		super(options);

		if (this.options.isWS) {
			this.ws = new WS(this.options);
			this.ws.on('connect', () => {
				this.wsReady = true;
			});
			this.ws.on('close', () => {
				this.wsReady = false;
			});
		}

		this.rest = new REST(this.options);
	}

	async Init() {
		let m = await this.GetMarket();
		if (m.MinTradeAmount) this.options.MinTradeAmount = m.MinTradeAmount;
		if (m.MinTradeValue) this.options.MinTradeValue = m.MinTradeValue;

		let t = await this.GetTicker();
		this.approxPrice = t.Last;
	}

	async Subscribe(Currency, BaseCurrency, type) {
		if (!this.options.isWS) throw new Error('is not websocket mode');
		if (['Depth', 'PublicTrades', 'Ticker'].indexOf(type) === -1) {
			throw new Error('unkown subscription type: ' + type);
		}

		if (type === 'Depth' && !this.options.onDepth) throw new Error('no onDepth callback');
		if (type === 'Ticker' && !this.options.onTicker) throw new Error('no onTicker callback');
		if (type === 'PublicTrades' && !this.options.onPublicTrades) throw new Error('no onPublicTrades callback');

		try {
			await this.ws.addSubscription(Currency, BaseCurrency, type);
		} catch (err) {
			console.error(this.GetName() + ` Subscribe got error:`, err);
			throw err;
		}
	}

	GetOrder(orderId, Currency, BaseCurrency) {
		return this.rest.GetOrder(orderId, Currency, BaseCurrency).then(data => {
			if (data && data.orderId) {
				return this._transform_order(data);
			} else {
				throw new Error('Order not found ' + orderId + ' ' + JSON.stringify(data));
			}
		});
	}

	_transform_order(data) {
		const STATUS = {
			'TRADING': 'Pending',
			'NEW': 'Pending',
			'FILLED': 'Closed',
			'CANCELED': 'Cancelled'
		};

		return {
			Id: data.orderId,
			Price: N.parse(data.price),
			AvgPrice: N.parse(data.price),
			Amount: N.parse(data.origQty || data.qty),
			DealAmount: N.parse(data.executedQty || data.qty),
			Status: STATUS[data.status] || 'Closed',
			Type: (data.side === 'BUY' || data.isBuyer) ? 'Buy' : 'Sell',
			Fee: 0,
			Time: N.parse(data.time),
			...this.rest._parse_symbol(data.symbol),
			Info: data
		};
	}

	CancelPendingOrders(Currency, BaseCurrency) {
		return this.GetOrders(Currency, BaseCurrency).then(async orders => {
			let ids = orders.map(o => o.Id);
			console.log('cancelling', ids);
			for (let id of ids)	{
				await this.rest.CancelOrder(id, Currency, BaseCurrency);
			}
			return true;
		});
	}

	GetOrders(Currency, BaseCurrency) {
		return this.rest.GetOrders(Currency, BaseCurrency).then(orders => {
			return orders.map(this._transform_order.bind(this));
		});
	}

	GetTrades(Currency, BaseCurrency) {
		return this.rest.GetTrades(Currency, BaseCurrency).then(orders => {
			return orders.map(o => {
				let no = this._transform_order(o);
				no.Id = o.id;
				return no;
			});
		}).then(trades => {
			return R.sort( R.ascend( R.prop('Time') ), trades);
		});
	}


}


module.exports = BINANCE;
