const WebSocket = require('../../lib/auto-reconnect-ws.js');
const N = require('precise-number');
const R = require('ramda');
const Events = require('events');
const debug = require('debug')('binance:ws');
const DelayTrigger = require('../../lib/delay-trigger.js');

class BINANCE_WS extends Events {

	constructor(options) {
		super();

		this.wsReady = false;
		this.alive = false;
		this.options = options;
		this.symbol = this.options.Currency.toLowerCase() + this.options.BaseCurrency.toLowerCase();
		this.trigger = new DelayTrigger(1000);

		this.trigger.on('connect', () => {
			this.connect();
		});

		this.lastPong = 0;

		this.pairs = {};
		this.pairs[this.symbol] = [this.options.Currency, this.options.BaseCurrency];

		//remember subscription commands
		this.subscriptionCommands = [];
		if (this.options.onDepth) {
			this.subscriptionCommands.push(`${this.symbol}@depth${this.options.DefaultDepthStep}`);
		}
		if (this.options.onTicker) {
			this.subscriptionCommands.push(`${this.symbol}@ticker`);
		}
		if (this.options.onPublicTrades) {
			this.subscriptionCommands.push(`${this.symbol}@aggTrade`);
		}

		this.trigger.trigger('connect');
	}

	connect() {
		let socketURL = `wss://stream.binance.com:9443/stream?streams=${this.subscriptionCommands.join('/')}`;
		if (socketURL === this.currentSocketURL) {
			console.log('trying to connect to the same socket url');
			return;
		}

		if (this.ws) {
			this.ws.close();
		}
		this.currentSocketURL = socketURL;

		console.log('new binance websocket connection', socketURL);
		this.ws = new WebSocket(socketURL, {
			pingInterval: 5000
		});

		this.ws.on('message', data => {
			try {
				debug('>>>', data);
				let message = JSON.parse(data);
				this.onMessage(message);
				this.triggerOnConnect();
			} catch (e) {
				console.log('huobipro message data error', e);
			}
		});

		this.ws.on('error', (err) => {
			console.log('huobipro websocket error', err);
		});

		this.ws.on('close', () => {
			this.wsReady = false;
			console.log('huobipro websocket closed');
			this.emit('close');
		});

		this.ws.on('pong', (ms) => {
			this.lastPong = new Date().getTime();
			if (this.options.onPong) this.options.onPong(ms);
		});

		this.ws.on('open', () => {
			this.triggerOnConnect();
		});
	}

	triggerOnConnect() {
		if (!this.wsReady) {
			this.wsReady = true;
			this.emit('connect');
		}
	}

	addSubscription(Currency, BaseCurrency, type) {
		let symbol = String(Currency + BaseCurrency).toLowerCase();
		console.log('add subscription', symbol, type);
		this.pairs[symbol] = [Currency, BaseCurrency];
		let sub = '';

		if (type === 'Depth') {
			sub = `${symbol}@depth${this.options.DefaultDepthStep}`;
		} else if (type === 'PublicTrades') {
			sub = `${symbol}@aggTrade`;
		} else if (type === 'Ticker') {
			sub = `${symbol}@ticker`;
		}
		this.subscriptionCommands.push(sub);
		this.trigger.trigger('connect');
	}

	onMessage(data)  {
		if (data && data.stream && data.data) {
			if (data.stream.match(/\@depth\d+$/)) {
				this.onDepth(data);
			} else if (data.stream.match(/\@ticker$/)) {
				this.onTicker(data);
			} else if (data.stream.match(/\@aggTrade$/)) {
				this.onPublicTrades(data);
			}
		} else {
			console.log('unknown websocket frame:', data);
		}
	}

	_parse_stream(stream) {
		let [symbol] = stream.split('@', 2);
		let [Currency, BaseCurrency] = this.pairs[symbol];
		return { Currency, BaseCurrency };
	}

	onDepth({stream, data}) {
		if (!data.asks || !data.bids) return;
		let asks = data.asks.map(pair => {
			return {
				Price: N.parse(pair[0]),
				Amount: N.parse(pair[1])
			};
		});

		let bids = data.bids.map(pair => {
			return {
				Price: N.parse(pair[0]),
				Amount: N.parse(pair[1])
			};
		});

		let depth = {
			Asks: R.sort( R.ascend( R.prop('Price') ), asks),
			Bids: R.sort( R.descend( R.prop('Price') ), bids),
			...this._parse_stream(stream),
			Time: Date.now(),
			Info: {
				lastUpdateId: data.lastUpdateId
			}
		};
		this.options.onDepth(depth);
	}

	onTicker({stream, data: ticker}) {
		if (!stream || !ticker) return;
		this.options.onTicker({
			Buy: N.parse(ticker.b),
			Sell: N.parse(ticker.a),
			High: N.parse(ticker.h),
			Last: N.parse(ticker.c),
			Low: N.parse(ticker.l),
			Volume: N.parse(ticker.v),
			Time: N.parse(ticker.E),
			...this._parse_stream(stream),
			Info: ticker
		});
	}

	onPublicTrades({stream, data: t}) {
		if (!stream || !t) return;

		t = {
			Amount: N.parse(t.q),
			Price: N.parse(t.p),
			Type: t.m ? 'Sell' : 'Buy',
			Time: Math.floor(t.T),
			Id: t.a,
			...this._parse_stream(stream)
		};
		this.options.onPublicTrades([t]);
	}

}


module.exports = BINANCE_WS;
