const REST = require('./rest.js');
const WS = require('./ws.js');
const N = require('precise-number');
const R = require('ramda');
const EXCHANGE = require('../exchange.js');
const { ok } = require('assert');

class BINANCE_FUTURE extends EXCHANGE {
	constructor(options) {
		options = Object.assign({
			Name: 'Binance',
			Fees: {
				Maker: 0.002,
				Taker: 0.002
			},
			RateLimit: 10,
			MinTradeAmount: 0.1,
			DefaultDepthStep: 20
		}, options);
		options.domain = options.domain || 'api.binance.com';
		super(options);

		if (this.options.isWS) {
			this.ws = new WS(this.options);
			this.ws.on('connect', () => {
				this.wsReady = true;
			});
			this.ws.on('close', () => {
				this.wsReady = false;
			});
		}

		this.rest = new REST(this.options);
	}

	async Init() {
		let m = await this.GetMarket();
		if (m.MinTradeAmount) this.options.MinTradeAmount = m.MinTradeAmount;
		if (m.MinTradeValue) this.options.MinTradeValue = m.MinTradeValue;

		let t = await this.GetTicker();
		this.approxPrice = t.Last;
	}

	async Subscribe(Currency, BaseCurrency, type) {
		if (!this.options.isWS) throw new Error('is not websocket mode');
		if (['Depth', 'PublicTrades', 'Ticker'].indexOf(type) === -1) {
			throw new Error('unkown subscription type: ' + type);
		}

		if (type === 'Depth' && !this.options.onDepth) throw new Error('no onDepth callback');
		if (type === 'Ticker' && !this.options.onTicker) throw new Error('no onTicker callback');
		if (type === 'PublicTrades' && !this.options.onPublicTrades) throw new Error('no onPublicTrades callback');

		try {
			await this.ws.addSubscription(Currency, BaseCurrency, type);
		} catch (err) {
			console.error(this.GetName() + ` Subscribe got error:`, err);
			throw err;
		}
	}

	GetOrder(orderId, Currency, BaseCurrency) {
		return this.rest.GetOrder(orderId, Currency, BaseCurrency).then(data => {
			if (data && data.orderId) {
				return this._transform_order(data);
			} else {
				throw new Error('Order not found ' + orderId + ' ' + JSON.stringify(data));
			}
		});
	}

	CancelOrder(orderId, Currency, BaseCurrency) {
		return this.rest.CancelOrder(orderId, Currency, BaseCurrency).then(data => {
			if (data && data.orderId) {
				return true;	
			} else {
				throw new Error('Cancel order error ' + orderId + ' ' + JSON.stringify(data));
			}
		}).catch(err => {
			if (err && err.code === -2011) {
				return true;
			} else {
				throw err;
			}
		});
	}

	CancelAndGetOrder(orderId, Currency, BaseCurrency) {
		return this.rest.CancelOrder(orderId, Currency, BaseCurrency).then(data => {
			if (data && data.orderId) {
				return this._transform_order(data);
			} else {
				throw new Error('Cancel order error ' + orderId + ' ' + JSON.stringify(data));
			}
		}).catch(err => {
			if (err && err.code === -2011) {
				return this.GetOrder(orderId, Currency, BaseCurrency);
			} else {
				throw err;
			}
		});
	}

	_transform_order(data) {
		const STATUS = {
			'TRADING': 'Pending',
			'NEW': 'Pending',
			'FILLED': 'Closed',
			'CANCELED': 'Cancelled',
			'EXPIRED': 'Cancelled'
		};

		let Price = N.parse(data.price);
		let AvgPrice = Price;
		if (data.cumQuote && data.executedQty && data.executedQty * 1 > 0) {
			AvgPrice = N(data.cumQuote).div(data.executedQty) * 1;
		}
		if (!Price && AvgPrice) Price = AvgPrice;
		return {
			Id: data.orderId,
			ClientId: data.clientOrderId,
			Price,
			AvgPrice,
			Amount: N.parse(data.origQty || data.qty),
			DealAmount: N.parse(data.executedQty || data.qty),
			Status: STATUS[data.status] || 'Closed',
			Type: (data.side === 'BUY' || data.isBuyer) ? 'Buy' : 'Sell',
			Fee: N.parse(data.commission || 0),
			Time: N.parse(data.time || data.updateTime || Date.now()),
			...this.rest._parse_symbol(data.symbol),
			Info: data
		};
	}

	CancelPendingOrders(Currency, BaseCurrency) {
		return this.GetOrders(Currency, BaseCurrency).then(async orders => {
			let ids = orders.map(o => o.Id);
			console.log('cancelling', ids);
			for (let id of ids)	{
				await this.rest.CancelOrder(id, Currency, BaseCurrency);
			}
			return true;
		});
	}

	GetOrders(Currency, BaseCurrency) {
		return this.rest.GetOrders(Currency, BaseCurrency).then(orders => {
			return orders.map(this._transform_order.bind(this));
		});
	}

	GetPosition(...args) {
		return this.rest.GetPosition(...args);
	}

	GetTrades(Currency, BaseCurrency, limit) {
		return this.rest.GetTrades(Currency, BaseCurrency, limit).then(orders => {
			return orders.map(o => {
				let no = this._transform_order(o);
				no.Id = o.id;
				no.OrderId = o.orderId;
				return no;
			});
		});
	}

	Trade(direction, price, amount, Currency, BaseCurrency, ContractType, options) {
		if (['Long', 'Short', 'CloseLong', 'CloseShort'].indexOf(direction) === -1) {
			throw new Error('unknown direction: ' + direction + ', direction should be one of these values: Long Short CloseLong CloseShort');
		}
		ok( amount > 0, 'amount should greater than 0');

		if (typeof Currency === 'object') {
			options = Currency;
			Currency = undefined;
		}
		if (typeof BaseCurrency === 'object') {
			options = BaseCurrency;
			BaseCurrency = undefined;
		}
		if (typeof ContractType === 'object') {
			options = ContractType;
			ContractType = undefined;
		}
		if (!options) options = {};
		options.ClientId = this.GetClientId(options.ClientId);
		
		console.log(this.GetName(), direction, price, amount, options);
		return this.rest.Trade(direction, price, amount, Currency, BaseCurrency, ContractType, options).then( r => {
			if (r && r.orderId) {
				return r.orderId;
			} else {
				console.log(r);
				throw new Error(r);
			}
		});
	}
}


module.exports = BINANCE_FUTURE;
