const fetch = require('node-fetch');
const N = require('precise-number');
const R = require('ramda');
const crypto = require('crypto');
const debug = require('debug')('binance:rest');
const ExError = require('../../lib/error');
const ErrorCode = require('../../lib/error-code');

const agent = require('../../lib/agent');

class BINANCE_REST {

	constructor(options) {
		this.key = options.Key;
		this.secret = options.Secret;
		this.symbol =  options.Currency.toUpperCase() + options.BaseCurrency.toUpperCase();
		this.options = options;
		this.Currency = options.Currency;
		this.BaseCurrency = options.BaseCurrency;

		if (!this.options.rateLimiter) throw 'No rateLimiter in options';
	}

	_getSymbol(Currency, BaseCurrency) {
		let c = Currency || this.options.Currency;
		let bc = BaseCurrency || this.options.BaseCurrency;
		return String(c + bc).toUpperCase();
	}

	_parse_symbol(ch, Currency, BaseCurrency) {
		if (!Currency) Currency = this.options.Currency;
		if (!BaseCurrency) BaseCurrency = this.options.BaseCurrency;

		let ms = this._split_symbol(ch);
		if (ms[0]) Currency = ms[0];
		if (ms[1]) BaseCurrency = ms[1];
		if (!ms[0] || !ms[1]) console.log('unknown symbol: ' + ch);
		return { Currency, BaseCurrency };
	}

	_split_symbol(symbol) {
		let reg = new RegExp('^([^\.]+?)(USD|USDT)$', 'i');
		let ms = String(symbol).match(reg);
		return [
			ms && ms[1] ? String(ms[1]).toUpperCase() : '',
			ms && ms[2] ? String(ms[2]).toUpperCase() : ''
		];
	}	

	async fetch(httpMethod, method, params, isPublic = false) {
		await this.options.rateLimiter.wait();
		if (!params) params = {};

		let sortedParams = R.compose(
			R.join('&'),
			R.sort(R.ascend(R.identity)),
			R.map(a => a[0] + '=' + encodeURIComponent(a[1])),
			R.toPairs
		)(params);

		let headers = {};

		if (!isPublic) {
			sortedParams += (sortedParams ? '&' : '') + 'timestamp=' + Date.now(); 
			let signString = sortedParams;
			let hash = crypto.createHmac('sha256', Buffer.from(this.secret, 'utf8')).update(signString).digest('hex');
			sortedParams += '&signature=' + encodeURIComponent(hash);
			headers['X-MBX-APIKEY'] = this.key;
		}

		let url = 'https://fapi.binance.com' + method;
		
		url += '?' + sortedParams;

		debug('<<<', httpMethod, url);

		return fetch(url, {
			method: httpMethod,
			timeout: isPublic ? 5000 : 10000,
			headers,
			agent: agent.https
		}).then(async res => {
			let s = await res.text();
			let status = res.status;
			if (status === 429) throw new ExError(ErrorCode.REQUEST_TOO_FAST, '429 request too fast');
			// if (status !== 200) throw new ExError(ErrorCode.BAD_RESPONSE_STATUS, `Binance response bad status(${status}): ${s}`);

			debug('>>>', status, s);
			let data = null;
			try {
				data = JSON.parse(s);
			} catch (err) {
				throw new ExError(ErrorCode.INVALID_JSON, 'BinanceRest can not parse json', err);
			}

			if (data && data.code && data.msg) {

				if (data.msg.match(/insufficient balance/)) {
					throw new ExError(ErrorCode.INSUFFICIENT_BALANCE, data.msg, data);
				}

				let err = new Error(data.code + ':' + data.msg);
				err.code = data.code;
				err.url = url;
				err.error_code = data.code;
				throw err;
			}
			return data;
		}).catch(err => {
			if (err.type === 'request-timeout') {
				throw new ExError(ErrorCode.REQUEST_TIMEOUT, `binance rest request (${params.method}) timeout`, err);
			}
			throw err;
		});
	}

	GetName() {
		return this.options.Name ? this.options.Name : 'Binance';
	}

	get(method, params) {
		return this.fetch('GET', method, params);
	}

	getPublic(method, params) {
		return this.fetch('GET', method, params, true);
	}

	post(method, params) {
		return this.fetch('POST', method, params);
	}

	error(...args) {
		args = args.map(a => {
			if (typeof a === 'object') a = JSON.stringify(a);
			return a;
		});
		let err = new Error(this.GetName() + ': ' + args.join(' '));
		throw err;
	}

	GetTicker(Currency, BaseCurrency) {
		return this.getPublic('/fapi/v1/ticker/bookTicker', {
			symbol: this._getSymbol(Currency, BaseCurrency)
		}).then(data => {
			return {
				Buy: N.parse(data.bidPrice),
				Sell: N.parse(data.askPrice),
				Last: N.parse(data.bidPrice),
				Time: Date.now(),
				...this._parse_symbol(data.symbol),
				Info: data
			};
		});
	}

	GetTickers() {
		return this.getPublic('/fapi/v1/ticker/bookTicker').then(data => {
			return data.filter(t => {
				return N(t.bidPrice).multiply(t.bidQty) > 0;
			}).map(d => {
				return {
					Buy: N.parse(d.bidPrice),
					Sell: N.parse(d.askPrice),
					Last: N.parse(d.bidPrice),
					Time: Date.now(),
					...this._parse_symbol(d.symbol),
					Info:d 
				};
			});
		});
	}

	GetMarkets() {
		return this.getPublic('/fapi/v1/exchangeInfo').then(data => {
			let bases = {};
			let markets = data.symbols.filter(m => {
				return m.status === 'TRADING';
			}).map(m => {
				bases[m.quoteAsset] = true;
				let re = {
					Currency: m.baseAsset,
					BaseCurrency: m.quoteAsset,
					ContractType: 'SWAP',
					Decimals: m.quotePrecision,
					StockDecimals: m.baseAssetPrecision,
					MinTradeAmount: N(1).div(Math.pow(10, m.baseAssetPrecision)) * 1,
					UnitValue: 1,
					UnitCurrency: m.baseAsset
				};

				let filters = this._parse_filters(m.filters);
				if (filters.MIN_NOTIONAL && filters.MIN_NOTIONAL.minNotional) {
					re.MinTradeValue = N.parse(filters.MIN_NOTIONAL.minNotional);
				}

				if (filters.LOT_SIZE && filters.LOT_SIZE.minQty) {
					re.MinTradeAmount = N.parse(filters.LOT_SIZE.minQty);
				}

				if (filters.LOT_SIZE && filters.LOT_SIZE.stepSize) {
					re.StockDecimals = this._countDecimals(N.parse(filters.LOT_SIZE.stepSize));
				}

				if (filters.PRICE_FILTER && filters.PRICE_FILTER.tickSize) {
					re.Decimals = this._countDecimals(N.parse(filters.PRICE_FILTER.tickSize));
				}

				re.Info = m;
				return re;
			});

			return markets;
		});
	}

	_countDecimals(ts) {
		let n = 10;
		for (let i = n; i > -10; i--) {
			let n1 = Math.pow(10, i);
			let n2 = Math.pow(10, i - 1);
			if (n1 >= ts && n2 < ts) {
				return -i;
			}
		}
	}

	_parse_filters(filters) {
		let re = {};
		filters.map(f => {
			re[f.filterType] = f;
		});
		return re;
	}

	GetDepth(Currency, BaseCurrency, limit = 50) {
		return this.getPublic('/fapi/v1/depth', {
			symbol: this._getSymbol(Currency, BaseCurrency),
			limit	
		}).then(data => {
			if (!data || !data.bids || !data.asks) throw new Error('get Binance ' + this.symbol + ' depth error ' + JSON.stringify(data));

			let asks = data.asks.map(r => {
				return {
					Price: N.parse(r[0]),
					Amount: N.parse(r[1])
				};
			});

			let bids = data.bids.map(r => {
				return {
					Price: N.parse(r[0]),
					Amount: N.parse(r[1])
				};
			});

			let Time = data && data.E ? N.parse(data.E) : Date.now();

			return Promise.resolve({
				Asks: R.sort( R.ascend( R.prop('Price') ), asks),
				Bids: R.sort( R.descend( R.prop('Price') ), bids),
				Info: {
					lastUpdateId: data.lastUpdateId,
					E: data.E,
					T: data.T,
					now: Date.now()
				},
				Time,
				Currency: Currency || this.options.Currency,
				BaseCurrency: BaseCurrency || this.options.BaseCurrency
			});
		});
	}

	GetTrades(Currency, BaseCurrency, limit = 500) {
		return this.get('/fapi/v1/userTrades', {
			symbol: this._getSymbol(Currency, BaseCurrency),
			limit
		}).then(data => {
			if (!data || !Array.isArray(data)) this.error('got broken trades data:' + JSON.stringify(data));
			return data;
		});
	}

	GetMarketDetail() {
		return this.get('/market/detail', {
			symbol: this.symbol
		}).then(data => {
			if (!data.tick) this.error('got broken market detail data:', data);
			return data.tick;
		});
	}

	async GetAccount() {
		return this.get('/fapi/v1/balance').then(data => {
			data = data[0];
			debug("binance future got balance:", data);
			let re = {
				Balance: N.parse(data.withdrawAvailable),
				FrozenBalance: N(data.balance).minus(data.withdrawAvailable) * 1,
				Stocks: 0,
				FrozenStocks:0,
				Value: N.parse(data.balance),
				Info: data
			};
			return re;
		});
	}

	async GetAccounts() {
		let a = await this.GetAccount();
		return [{
			Currency: 'USDT',
			Free: a.Balance,
			Frozen: a.FrozenBalance,
			Risk: 0,
			Value: a.Value,
			Info: a.Info
		}];
	}

	async GetPosition(Currency, BaseCurrency) {
		if (!Currency) Currency = this.options.Currency;
		if (!BaseCurrency) BaseCurrency = this.options.BaseCurrency;
		let symbol = this._getSymbol(Currency, BaseCurrency);
		return this.get('/fapi/v1/positionRisk', {
			symbol
		}).then(data => {
			data = data.filter(d => d.symbol === symbol).shift();
			let re = {
				MarginLevel: N.parse(data.leverage),
				Amount: N.parse(data.positionAmt),
				FrozenAmount: 0,
				Price: N.parse(data.markPrice),
				Profit: N.parse(data.unRealizedProfit),
				Info: data
			};
			re.Type = 'United';
			return re;
		});
	}

	GetOrder(orderId, Currency, BaseCurrency) {
		return this.get('/fapi/v1/order', {
			symbol: this._getSymbol(Currency, BaseCurrency),
			orderId
		});
	}

	GetOrders(Currency, BaseCurrency) {
		return this.get('/fapi/v1/openOrders', {
			symbol: this._getSymbol(Currency, BaseCurrency)
		}).then(r => {
			if (!r || !Array.isArray(r)) this.error('can not get orders:', r);
			return r;
		});
	}

	GetOpenOrders() {
		return this.get('/v1/order/openOrders', {
			size: 500
		}).then(r => {
			if (!r.data) this.error('can not get orders:', r);
			return r.data;
		});
	}

	GetPublicTrades(Currency, BaseCurrency, size = 500) {
		return this.getPublic('/fapi/v1/aggTrades', {
			symbol: this._getSymbol(Currency, BaseCurrency),
			limit: size
		}).then(data => {
			return data.map(t => {
				return {
					Id: t.a,
					Time: N.parse(t.T),
					Price: N.parse(t.p),
					Amount: N.parse(t.q),
					Type: t.m ? 'Sell' : 'Buy',
					Currency: Currency || this.options.Currency,
					BaseCurrency: BaseCurrency || this.options.BaseCurrency
				};
			});
		});
	}

	CancelOrder(orderId, Currency, BaseCurrency) {
		return this.fetch('DELETE', '/fapi/v1/order', {
			symbol: this._getSymbol(Currency, BaseCurrency),
			orderId
		});
	}

	CancelOrders(ids) {
		if (!ids || ids.length === 0) return true;
		return this.post('/v1/order/orders/batchcancel', {
			'order-ids': ids
		}).then(r => {
			if (r && r.data) {
				console.log('cancel order result:', r.data);
			}
			return true;
		});
	}

	async Trade(direction, price, amount, Currency, BaseCurrency, ContractType, options) {
		if (direction === 'CloseLong' || direction === 'Short') direction = 'SELL';
		if (direction === 'CloseShort' || direction === 'Long') direction = 'BUY';
		let params = {
			symbol: this._getSymbol(Currency, BaseCurrency),
			side: direction,
			type: 'LIMIT',
			quantity: amount + '',
			price: price + '',
			timeInForce: this.getOrderType(options.Type),
		};

		if (options.ClientId) {
			params.newClientOrderId = options.ClientId;
		}

		if (price === -1 || options.Type === 'MARKET') {
			params.type = 'MARKET';
			delete params.price;
			delete params.timeInForce;
		}

		return this.fetch('POST', `/fapi/v1/order`, params);
	}

	getOrderType(Type) {
		/*
		GTC - Good Till Cancel 成交为止
		IOC - Immediate or Cancel 无法立即成交(吃单)的部分就撤销
		FOK - Fill or Kill 无法全部立即成交就撤销
		GTX - Good Till Crossing 无法成为挂单方就撤销
		 */
		let types = ['GTC', 'IOC', 'FOK', 'GTX'];
		if (Type === 'PO') return 'GTX';
		if (!Type || types.indexOf(Type) === -1) return this.options.MakerMode ? 'GTX' : 'GTC';
		return Type;
	}

	GetRecords(minutes = 1, Currency = '', BaseCurrency = '', limit = 200) {
		let period = '';
		if (minutes === 1) period = '1m';
		if (minutes === 3) period = '3m';
		if (minutes === 5) period = '5m';
		if (minutes === 15) period = '15m';
		if (minutes === 30) period = '30m';
		if (minutes === 60) period = '1h';
		if (minutes === 120) period = '2h';
		if (minutes === 240) period = '4h';
		if (minutes === 360) period = '6h';
		if (minutes === 480) period = '8h';
		if (minutes === 720) period = '12h';
		if (minutes === 1440) period = '1d';
		if (minutes === 4320) period = '3d';
		if (minutes === 10080) period = '1w';
		if (minutes === 43200) period = '1M';
		if (!period) return Promise.reject(new Error('unsupported binance kline period: ' + minutes + ' minutes'));

		return this.getPublic('/fapi/v1/klines', {
			symbol: this._getSymbol(Currency, BaseCurrency),
			interval: period,
			limit
		}).then(data => {
			return data.map(k => {
				return {
					Open: N.parse(k[1]),
					High: N.parse(k[2]),
					Low: N.parse(k[3]),
					Close: N.parse(k[4]),
					Volume: N.parse(k[5]),
					CurrencyVolume: N.parse(k[7]),
					Time: N.parse(k[6])
				};
			});
		});
	}


}

module.exports = BINANCE_REST;
