const N = require('precise-number');
const { ok } = require('assert');
const REST = require('./rest.js');
const WS = require('./ws.js');
const EventEmitter = require('events');
const EXCHANGE = require('../exchange.js');
const ExError = require('../../lib/error');
const ErrorCode = require('../../lib/error-code');
const wait = require('delay');
const color = require('turbocolor');

class BITFLYER_FX extends EXCHANGE {
	constructor(options) {

		options = Object.assign({
			Name: 'BITFLYER_FX',
			Fees: {
				Maker: 0,
				Taker: 0
			},
			RateLimit: 3,
			Decimals: 0,
			StockDecimals: 8,
			MinTradeAmount: 0.1,
			BaseCurrency: 'JPY',
			Currency: 'BTC',
			MarginLevel: 3,
			WSTimeout: 20000,
			SnapshotMode: false,
			RateLimit: 5,
			RateLimitInterval: 3000,
			ThrowWhenRateLimited: true,
		}, options);
		super(options);

		options.BaseCurrency = 'JPY';
		options.Currency = 'BTC';
		
		this.symbol = 'FX_' + options.Currency + '_' + options.BaseCurrency;
		this.options.ContractType = this.symbol;

		this.fee = {
			Maker: 0,
			Taker: 0
		};

		this.rest = new REST(this.options);

		this.orderbook = null;

		this.lastDepthTime = 0;
		this.lastDepth = null;

		this.events = new EventEmitter();

		if (this.options.isWS) {
			this.ws = new WS(this.options);
			this.ws.on('ready', () => {
				this.wsReady = true;
			});
		}
	}

	GetDepth() {
		if (this.isWS && this.wsReady) {
			if (Date.now() - this.lastDepthTime < 10 && this.lastDepth) {
				return Promise.resolve(this.lastDepth);
			}
			return new Promise(done => {
				this.events.once('depth', d => done(d));
			});
		} else {
			return this.rest.GetDepth();
		}
	}

	GetTicker() {
		/*
		{ product_code: 'BTC_JPY',
		  timestamp: '2017-10-10T13:49:30.187',
		  tick_id: 13257531,
		  best_bid: 537178,
		  best_ask: 537345,
		  best_bid_size: 2e-7,
		  best_ask_size: 0.00976006,
		  total_bid_depth: 4636.9179559,
		  total_ask_depth: 1793.25003469,
		  ltp: 537346,
		  volume: 203079.27801949,
		  volume_by_product: 23606.46619995 }
		 */
		return this.rest.GetTicker().then(data => {
			return {
				Buy: N.parse(data.best_bid),
				Sell: N.parse(data.best_ask),
				Last: N(data.best_bid).add(data.best_ask).div(2) * 1,
				Time: new Date(data.timestamp + 'Z').getTime(),
				High: N.parse(data.best_ask),
				Low: N.parse(data.best_bid),
				Volume: N.parse(data.volume_by_product),
				Currency: this.options.Currency,
				BaseCurrency: this.options.BaseCurrency,
				ContractType: this.options.ContractType,
				Info: data
			};
		});
	}

	GetCollateral() {
		return this.rest.GetCollateral();
	}

	GetPosition() {
		/*
		[ { product_code: 'FX_BTC_JPY',
		    side: 'SELL',
		    price: 748756,
		    size: 0.01,
		    commission: 0,
		    swap_point_accumulate: 0,
		    require_collateral: 2495.8533333333335,
		    open_date: '2018-06-20T09:10:42.083',
		    leverage: 3,
		    pnl: 3.19,
		    sfd: 0 } ]
		 [ { product_code: 'FX_BTC_JPY',
		     side: 'BUY',
		     price: 748324,
		     size: 0.01,
		     commission: 0,
		     swap_point_accumulate: 0,
		     require_collateral: 2494.4133333333334,
		     open_date: '2018-06-20T09:13:42.577',
		     leverage: 3,
		     pnl: -1.06,
		     sfd: 0 } ]
		 */
		return this.rest.GetPosition().then(positions => {
			if (!positions || !positions.map) {
				console.log('bad positions', positions);
				throw new Error('bad position data' + JSON.stringify(positions));
			}
			return positions.map(p => {
				let re = {
					Amount: N.parse(p.size),
					MarginLevel: N.parse(p.leverage),
					FrozenAmount: 0,
					Price: N.parse(p.price),
					Type: p.side === 'BUY' ? 'Long' : 'Short',
					ContractType: p.product_code,
					Currency: this.options.Currency,
					BaseCurrency: this.options.BaseCurrency,
					Info: p
				};
				return re;
			});
		}).then(positions => {
			if (positions.length <= 1) return positions;
			let longPosition = this._merge_position(positions, 'Long');
			let shortPosition = this._merge_position(positions, 'Short');
			let re = [];
			if (longPosition) re.push(longPosition);
			if (shortPosition) re.push(shortPosition);
			return re;
		});
	}

	async GetNetPosition() {
		let positions = await this.GetPosition();
		let re = {
			Amount: 0,
			ContractType: this.options.ContractType,
			Price: 0,
			MarginLevel: 0,
			Currency: this.options.Currency,
			BaseCurrency: this.options.BaseCurrency
		};

		positions.map( p => {
			if (p.Type === 'Long') {
				re.Amount = p.Amount;
				re.Price = p.Price;
				re.MarginLevel = p.MarginLevel;
			} else {
				re.Amount = -1 * p.Amount;
				re.Price = p.Price;
				re.MarginLevel = p.MarginLevel;
			}
		});

		return re;
	}

	_merge_position(positions, type) {
		let arr = positions.filter(p => p.Type === type);
		if (arr.length === 1) return arr[0];
		if (arr.length === 0) return null;
		let position = {
			Type: type,
			Amount: 0,
			MarginLevel: 0,
			FrozenAmount: 0,
			Price: 0,
			ContractType: '',
			Currency: this.options.Currency,
			BaseCurrency: this.options.BaseCurrency,
			Info: []
		};
		arr.map(p => {
			position.Amount = N(p.Amount).add(position.Amount) * 1;
			position.Price = N(p.Price).multiply(p.Amount).add(position.Price) * 1;
			position.MarginLevel = p.MarginLevel;
			position.ContractType = p.ContractType;
			position.Info.push(p);
		});

		position.Price = N(position.Price).div(position.Amount).floor(0);

		return position;
	}

	async GetAccount() {
		/*
		{ collateral: 48918.872,
		  open_position_pnl: 0,
		  require_collateral: 0,
		  keep_rate: 0 }
		 */
		let info = await this.rest.GetCollateral();
		let total = info.collateral || 0;
		let totalBalance = total;//Math.floor(total / 0.8);
		let usedBalance = Math.floor(info.require_collateral);
		let pnl = N.parse(info.open_position_pnl || 0);

		return {
			Balance: Math.floor(totalBalance - usedBalance),
			FrozenBalance: usedBalance,
			Value: Math.floor(totalBalance + pnl),
			Stocks: 0,
			FrozenStocks: 0,
			MarginLevel: info.keep_rate > 0 ? (N(1).div(info.keep_rate).multiply(this.options.MarginLevel) * 1) : 0,
			Currency: this.options.Currency,
			BaseCurrency: this.options.BaseCurrency,
			ContractType: this.options.ContractType,
			Info: info
		};
	}

	GetOrders() {
		/*
		[ { id: 0,
		    child_order_id: 'JFX20180620-092321-465119F',
		    product_code: 'FX_BTC_JPY',
		    side: 'SELL',
		    child_order_type: 'LIMIT',
		    price: 760550,
		    average_price: 0,
		    size: 0.01,
		    child_order_state: 'ACTIVE',
		    expire_date: '2018-07-20T09:23:20',
		    child_order_date: '2018-06-20T09:23:20',
		    child_order_acceptance_id: 'JRF20180620-092320-723852',
		    outstanding_size: 0.01,
		    cancel_size: 0,
		    executed_size: 0,
		    total_commission: 0 } ]
		 */
		return this.rest.GetOrders().then(orders => orders.map(o => this._transform_order(o)));
	}

	_transform_order(o) {
		return {
			Id: o.child_order_acceptance_id,
			Price: N.parse(o.price || o.average_price),
			AvgPrice: N.parse(o.average_price),
			Amount: N.parse(o.size),
			DealAmount: N.parse(o.executed_size),
			Type: o.side === 'BUY' ? 'Long' : 'Short',
			Time: new Date(o.child_order_date + '.000Z').getTime(),
			Status: this._order_status(o.child_order_state),
			ContractType: o.product_code,
			Currency: this.options.Currency,
			BaseCurrency: this.options.BaseCurrency,
			Info: o
		};
	}

	GetOrder(orderId) {
		return this.rest.GetOrders(orderId).then(orders => orders.map(o => this._transform_order(o))).then(orders => {
			if (!orders || orders.length === 0) {
				throw new ExError(ErrorCode.ORDER_NOT_FOUND, `order(${orderId}) not found`);
			}

			if (orders.length === 1) {
				return orders[0];
			} else {
				console.error(orders);
				throw new ExError(ErrorCode.UNKNOWN_ERROR, 'Bitflyer FX GetOrder returns many orders:' + JSON.stringify(orders));
			}
		});
	}

	_order_status(status) {
		/*
		ACTIVE: Return open orders
		COMPLETED: Return fully completed orders
		CANCELED: Return orders that have been cancelled by the customer
		EXPIRED: Return order that have been cancelled due to expiry
		REJECTED: Return failed orders
		 */
		switch (status) {
				case 'ACTIVE': return 'Pending';
				case 'COMPLETED': return 'Closed';
				case 'CANCELED': return 'Cancelled';
				case 'REJECTED': return 'Cancelled';
				case 'EXPIRED': return 'Cancelled';
				default: return status;
		}
	}

	Trade(direction, price, amount, options = {}) {
		if (!options) options = {};
		if (['Long', 'Short', 'CloseLong', 'CloseShort'].indexOf(direction) === -1) {
			throw new Error('unknown direction: ' + direction + ', direction should be one of these values: Long Short CloseLong CloseShort');
		}

		ok( amount > 0, 'amount should greater than 0');

		console.log(this.GetName(), direction, price, amount, options);
		return this.rest.Trade(direction, price, amount, options);
	}

	CancelOrder(orderId) {
		return this.rest.CancelOrder(orderId);
	}

	CancelPendingOrders() {
		console.log(this.GetName() + ' cancelling pending orders...');
		return this.rest.CancelPendingOrders();
	}

	GetMarket() {
		return {
			Currency: 'BTC',
			BaseCurrency: 'JPY',
			Decimals: 0,
			StockDecimals: 8,
			MinTradeAmount: 0.01
		};
	}

	GetMarkets() {
		return [ this.GetMarket() ];
	}
}


module.exports = BITFLYER_FX;
