const N = require('precise-number');
const R = require('ramda');
const moment = require('moment');
const Events = require('events');
const OrderBook = require('../../lib/orderbook.js');
const io = require('socket.io-client');
const wait = require('delay');

class BITFLYER_FX_WS extends Events {
	constructor(options) {
		super();
		this.options = options;

		this.symbol = 'FX_' + options.Currency + '_' + options.BaseCurrency;

		this.orderbook = null;

		this.lastDepthTime = 0;
		this.lastMessageTime = 0;
		this.gapHistory = [];
		this.lastDepth = null;

		this.DepthSize = this.options.WSDepthSize || 200;

		this.initiateWS();

		setInterval(() => {
			if (Date.now() - this.lastDepthTime >= 60000) {
				console.error('websocket dead, reconnecting ... ');
				this.initiateWS();
			}
		}, 60000);
	}

	initiateWS() {
		if (this.ws) {
			try {
				this.orderbook = null;
				this.ws.close();
			} catch (err) {
				console.error('close old ws error', err);
			}
		}

		try {

			this.wsReady = false;

			this.ws = io("https://io.lightstream.bitflyer.com", {
				transports: ["websocket"],
				reconnection: true,
				timeout: 10000
			});

			let snapshotChannel = "lightning_board_snapshot_" + this.symbol;
			let depthChannel = "lightning_board_" + this.symbol;
			let executionChannel = 'lightning_executions_' + this.symbol;
			let spotExecutionChannle = 'lightning_executions_BTC_JPY';

			this.ws.on("connect", () => {
				console.log('bitflyer_fx ws on connect');

				if (this.options.onDepth) {
					this.ws.emit("subscribe", snapshotChannel);
					if (!this.options.SnapshotMode) {
						this.ws.emit("subscribe", depthChannel);
					}
				}

				if (this.options.onPublicTrades) this.ws.emit('subscribe', executionChannel);
				if (this.options.onSpotPublicTrades) this.ws.emit('subscribe', spotExecutionChannle);
			});

			this.ws.on('connect_error', (err) => {
				console.warn('bitflyer websocket connect error', err);
			});

			this.ws.on('connect_timeout', (err) => {
				console.warn('bitflyer websocket connect timeout', err);
			});

			this.ws.on('error', (err) => {
				console.warn('bitflyer websocket connect error', err);
			});

			this.ws.on('disconnect', (err) => {
				console.warn('bitflyer websocket disconnect', err);
			});

			this.ws.on('reconnect', (times) => {
				console.warn('bitflyer websocket reconnect', times);
			});

			this.ws.on('reconnect_attempt', (times) => {
				console.warn('bitflyer websocket reconnect attempt', times);
			});

			this.ws.on('reconnecting', (times) => {
				console.warn('bitflyer websocket reconnecting', times);
			});

			this.ws.on('reconnect_error', (err) => {
				console.warn('bitflyer websocket reconnect error', err);
			});

			this.ws.on('reconnect_failed', (err) => {
				console.warn('bitflyer websocket reconnect failed', err);
			});

			this.ws.on('pong', (ms) => {
				if (this.options.onPong) this.options.onPong(ms);
			});

			this.ws.on(snapshotChannel, notify => {
				this.lastMessageTime = Date.now();
				this.ready();	
				this.buildOrderBook(notify);
			});

			this.ws.on(depthChannel, notify => {
				this.lastMessageTime = Date.now();
				this.ready();	
				this.updateOrderBook(notify);
			});

			this.ws.on(executionChannel, notify => {
				this.lastMessageTime = Date.now();
				this.ready();	
				this.onPublicTrades(notify);
			});

			this.ws.on(spotExecutionChannle, notify => {
				this.lastMessageTime = Date.now();
				this.ready();	
				this.onSpotPublicTrades(notify);
			});
		} catch (err) {
			console.error("new websocket error", err);
		}
	}

	ready() {
		this.wsReady = true;
		this.emit('ready');
	}

	onPublicTrades(trades) {
		if (!trades || trades.length <= 0) return;

		trades = trades.map(t => ({
			Amount: N.parse(t.size),
			Price: N.parse(t.price),
			Type: t.side === 'BUY' ? 'Buy' : 'Sell',
			Time: new Date(t.exec_date).getTime(),
			Id: t.id,
			Currency: this.options.Currency,
			BaseCurrency: this.options.BaseCurrency,
			// Info: t
		}));
		if (this.options.onPublicTrades) {
			this.options.onPublicTrades(trades);
		}
	}

	onSpotPublicTrades(trades) {
		if (!trades || trades.length <= 0) return;

		trades = trades.map(t => ({
			Amount: N.parse(t.size),
			Price: N.parse(t.price),
			Type: t.side === 'BUY' ? 'Buy' : 'Sell',
			Time: new Date(t.exec_date).getTime(),
			Id: t.id,
			Currency: 'BTC',
			BaseCurrency: 'JPY'
		}));
		if (this.options.onSpotPublicTrades) {
			this.options.onSpotPublicTrades(trades);
		}
	}

	async buildOrderBook(depth) {
		if (!depth) return;
		this.lastDepthTime = Date.now();
		let data = { Asks:[], Bids:[]};
		if (depth.asks && depth.asks.length > 0) {
			data.Asks = depth.asks.map(r => {
				return {
					Price: N.parse(r.price),
					Amount: N.parse(r.size)
				};
			});
		}
		if (depth.bids && depth.bids.length > 0) {
			data.Bids = depth.bids.map(r => {
				return {
					Price: N.parse(r.price),
					Amount: N.parse(r.size)
				};
			});
		}
		if (!this.orderbook) this.orderbook = new OrderBook();
		this.orderbook.build(data);
		this.onDepthData();
	}

	updateOrderBook(data) {
		this.lastDepthTime = Date.now();
		if (!this.orderbook) return;

		let Asks = [], Bids = [];
		if (data.bids && data.bids.length > 0) {
			Bids = data.bids.map(d => {
				return {
					Price: N.parse(d.price),
					Amount: N.parse(d.size)
				};
			});
			Bids = Bids.filter(e => e.Price > 0);
			Bids = R.sort(R.descend(R.prop('Price')))(Bids);
		}

		if (data.asks && data.asks.length > 0) {
			Asks = data.asks.map(d => {
				return {
					Price: N.parse(d.price),
					Amount: N.parse(d.size)
				};
			});
			Asks = Asks.filter(e => e.Price > 0);
			Asks = R.sort(R.ascend(R.prop('Price')))(Asks);
		}

		this.orderbook.update(Asks, Bids);
		this.onDepthData();
	}

	onDepthData() {
		let depth = {
			...this.orderbook.get(),
			Currency: this.options.Currency,
			BaseCurrency: this.options.BaseCurrency,
			ContractType: this.options.ContractType
		};

		if (typeof this.options.onDepth === 'function') {
			if (depth && depth.Asks && depth.Bids && depth.Asks.length > 0 && depth.Bids.length > 0) {
				depth.Asks = depth.Asks.slice(0, this.DepthSize);
				depth.Bids = depth.Bids.slice(0, this.DepthSize);
				this.options.onDepth(depth, this.lastMessageTime);
			}
		}

		this.lastDepth = depth;	
	}

	async watchGap() {
		while (true) {
			if (this.lastDepth) {
				let {Asks, Bids} = this.lastDepth;
				let gap = (Asks[0].Price - Bids[0].Price) / Bids[0].Price;
				this.gapHistory.push(gap);
				this.gapHistory = this.gapHistory.slice(-1000);
				let ma = this.gapHistory.reduce((a, c) => { return a + c; }, 0) / this.gapHistory.length;

			}
			await wait(500);
		}
	}

}


module.exports = BITFLYER_FX_WS;
