const N = require('precise-number');
const assert = require('better-assert');
const R = require('ramda');
const moment = require('moment');
const REST = require('./rest.js');
const WS = require('./ws.js');
const EventEmitter = require('events');
const EXCHANGE = require('../exchange.js');
const ExError = require('../../lib/error');
const ErrorCode = require('../../lib/error-code');

class BITMEX extends EXCHANGE {
	constructor(options) {

		options = Object.assign({
			Name: 'BITMEX',
			Fees: {
				Maker: -0.00025,
				Taker: 0.00075
			},
			RateLimit: 3,
			Decimals: 1,
			StockDecimals: 0,
			MinTradeAmount: 1,
			BaseCurrency: 'USD',
			Currency: 'XBT',
			MarginLevel: 100,
			RateLimit: 2,
			RateLimitInterval: 1000
		}, options);
		super(options);

		this.symbol = options.Currency + options.BaseCurrency;
		this.options.ContractType = this.symbol;

		this.rest = new REST(this.options);

		this.orderbook = null;

		this.lastDepthTime = 0;
		this.lastDepth = null;

		this.events = new EventEmitter();

		if (this.options.isWS) {
			this.ws = new WS(this.options);
		}
	}

	GetDepth() {
		if (this.isWS && this.wsReady) {
			if (Date.now() - this.lastDepthTime < 10 && this.lastDepth) {
				return Promise.resolve(this.lastDepth);
			}
			return new Promise(done => {
				this.events.once('depth', d => done(d));
			});
		} else {
			return this.rest.GetDepth();
		}
	}

	GetTicker() {
		return this.rest.GetTicker().then(data => {
			return {
				Buy: N.parse(data.best_bid),
				Sell: N.parse(data.best_ask),
				Last: N(data.best_bid).add(data.best_ask).div(2) * 1,
				Time: moment(data.timestamp).format('x') * 1,
				High: N.parse(data.best_ask),
				Low: N.parse(data.best_bid),
				Volume: N.parse(data.volume_by_product),
				Currency: this.options.Currency,
				BaseCurrency: this.options.BaseCurrency,
				ContractType: this.options.ContractType,
				Info: data
			};
		});
	}

	GetPosition(Currency, BaseCurrency) {
		return this.rest.GetPosition(Currency, BaseCurrency);
	}

	GetAccount() {
		return this.rest.GetAccount();
	}

	GetOrders(Currency, BaseCurrency) {
		return this.rest.GetOrders(Currency, BaseCurrency).then(orders => orders.map(o => this._transform_order(o)));
	}

	_transform_order(o) {
		return {
			Id: o.orderID,
			Price: N.parse(o.price),
			Amount: N.parse(o.orderQty),
			DealAmount: N.parse(o.cumQty),
			StopPx: o.stopPx ? N.parse(o.stopPx) : 0,
			Type: o.side === 'Buy' ? 'Long' : 'Short',
			Time: moment(o.timestamp).format('x') * 1,
			Status: this._order_status(o.ordStatus),
			ContractType: '--',
			Currency: o.settlCurrency.toUpperCase(),
			BaseCurrency: o.currency,
			Info: o
		};
	}

	GetOrder(orderId) {
		return this.rest.GetOrder(orderId).then(orders => orders.map(o => this._transform_order(o))).then(orders => {
			if (!orders || orders.length === 0) {
				throw new ExError(ErrorCode.ORDER_NOT_FOUND, `order(${orderId}) not found`);
			}

			if (orders.length === 1) {
				return orders[0];
			} else {
				console.error(orders);
				throw new ExError(ErrorCode.UNKNOWN_ERROR, 'Bitflyer FX GetOrder returns many orders:' + JSON.stringify(orders));
			}
		});
	}

	_order_status(status) {
		/*
		let arr = {
			'5':'Cancelled',
			'1': 'Pending',
			'2': 'Pending',
			'3': 'Pending',
			'4': 'Partial',
			'6': 'Closed',
			'7': 'Cancelled',
			'11': 'Pending'
		};
		 */
		 const statuses = {
            'New': 'Pending',
            'PartiallyFilled': 'Partial',
            'Filled': 'Closed',
            'DoneForDay': 'Pending',
            'Canceled': 'Cancelled',
            'PendingCancel': 'Pending',
            'PendingNew': 'Pending',
            'Rejected': 'Cancelled',
            'Expired': 'Cancelled',
            'Stopped': 'Pending',
            'Untriggered': 'Pending',
            'Triggered': 'Pending',
        };

        return statuses[status] || srtatus;
 
	}

	Trade( direction , price, amount , Currency, BaseCurrency) {
		amount = N(amount).floor(this.options.StockDecimals);
		if (price !== -1) price = N(price).floor(this.options.Decimals);
		return this.rest.Trade(direction, price, amount );
	}

	CancelOrder(orderId) {
		return this.rest.CancelOrder(orderId)
		.then( o =>{
			return true;
		})
		.catch( err => {
			return this.GetOrder(orderId).then(o=>{
				// console.log(o);
				if( o.Status === 'Cancelled' || o.Status === 'Closed' ){
					return true;
				}else{
					console.error('cancel order got error:', JSON.stringify(o, null, '\t'));
					return false;
				}
			})
		});
	}

	CancelPendingOrders(Currency, BaseCurrency) {
		console.log(this.GetName() + ' cancelling pending orders...');
		return this.rest.CancelPendingOrders(Currency, BaseCurrency);
	}

	CloseAllPositions(Currency, BaseCurrency) {
		console.log(this.GetName() + ' closing positions...');
		return this.rest.CloseAllPositions(Currency, BaseCurrency);
	}

}


module.exports = BITMEX;
