const fetch = require('node-fetch');
const N = require('precise-number');
const R = require('ramda');
const debug = require('debug')('bitmex:rest');
const crypto = require('crypto');

class Exchange {

	constructor(options) {
		this.key = options.Key;
		this.secret = options.Secret;
		this.symbol = options.Currency.toUpperCase() + options.BaseCurrency.toUpperCase();
		this.options = options;

		// this.bitmex = new BitMexPlus({
		// 	apiKeyID: this.key,
		// 	apiKeySecret: this.secret
		// });
	}

	async fetch(path, api = 'public', method = 'GET', params = {}, headers = undefined, body = undefined) {
		
		let info = this.sign(path, api, method, params, headers, body);	
		let options = {
			method,
			headers: info.headers,
			body: info.body
		};

		return fetch(info.url, options).then(res => res.text()).then(r => {
			debug(`${path} returns: `, r);
			try {
				r = JSON.parse(r);
			} catch (err) {
				throw new Error(`parse json response of ${path} error: ${err.message}`);
			}
			if (r && r.error) throw new Error(JSON.stringify(r.error));
			return r;
		});
	}

	urlencode(params) {
		let re = [];
		Object.keys(params).map(k => {
			if (typeof params[k] === 'object') params[k] = JSON.stringify(params[k]);
			re.push(`${k}=${encodeURIComponent(params[k])}`);
		});
		return re.join('&');
	}

	sign (path, api = 'public', method = 'GET', params = {}, headers = undefined, body = undefined) {
		let query = '/api/v1/' + path;
		if (method !== 'PUT' && Object.keys (params).length) query += '?' + this.urlencode(params);
		let url = 'https://www.bitmex.com' + query;
		if (api === 'private') {
			let nonce = Date.now() + '';
			let auth = method + query + nonce;
			if (method === 'POST' || method === 'PUT') {
				if (Object.keys (params).length) {
					body = JSON.stringify(params) ; //this.json (params);
					auth += body;
				}
			}
			headers = {
				'Content-Type': 'application/json',
				'api-nonce': nonce,
				'api-key': this.key,
				'api-signature': crypto.createHmac('sha256', this.secret).update(auth).digest('hex')
			};
		}
		return { url, method, body, headers };
	}

	GetTicker() {
		return this.fetch('quote/bucketed', 'private', 'GET', {
			symbol: this.symbol,
			binSize: '1d',
			partial: true,
			count: 1,
			reverse: true
		}, true).then(d => {
			if (!d || d.length !== 1) throw new Error('bitmex get ticker error:' + JSON.stringify(d));
			return d[0];
		}).then(t => {
			return {
				Buy: N.parse(t.bidPrice),
				Sell: N.parse(t.askPrice),
				BuyAmount: N.parse(t.bidSize),
				SellAmount: N.parse(t.askSize),
				High: N.parse(t.askPrice),
				Low: N.parse(t.bidPrice),
				Volume: 0,
				Time: Date.now(),
				Info: t
			};
		});
	}

	GetAccount() {
		return this.fetch('user/margin', 'private', 'GET', {}).then(data=>{
			let re = {
				Balance: 0,
				FrozenBalance: 0,
				Stocks: N(data.marginBalance) / 100000000,
				FrozenStocks: N(data.marginBalance).minus(data.availableMargin) / 100000000,
				Value: N(data.availableMargin) / 100000000,
				// Risk: N(info.risk_rate),
				Currency: 'XBT',
				BaseCurrency: 'USD',
				Info: data
			};
			return re;
		})
	}

	GetPosition(Currency, BaseCurrency) {
		let symbol = ( Currency && BaseCurrency) ? Currency + BaseCurrency : this.symbol;
		return this.fetch('position', 'private', 'GET', {
			filter: { symbol }
		}).then(positions=>{
			let re = [];
			positions.map(data=>{
				re.push({
					MarginLevel: data.leverage,
					Amount: Math.abs(N.parse(data.currentQty)),
					AmountValue: Math.abs(N.parse(data.homeNotional)),
					FrozenAmount: N.parse(data.realisedCost / 100000000),
					Price: N.parse(data.avgCostPrice),
					Profit: N.parse(data.unrealisedRoePcnt),
					Type: data.currentQty > 0 ? 'Long' : 'Short',
					ContractType: '--',
					Currency,
					BaseCurrency: 'USD',
					Info: data
				});
			});
			return re;
		});
	}

	GetOrders(Currency, BaseCurrency) {
		let symbol = ( Currency && BaseCurrency) ? Currency + BaseCurrency : this.symbol;
		return this.fetch('order', 'private', 'GET', {
			symbol: symbol,
			filter: { open: true }
		});	
	}

	GetOrder(orderId) {
		return this.fetch('order', 'private', 'GET', {
			filter:{orderID : orderId}
		});	
	}

	CancelOrder(orderId) {
		return this.fetch('order', 'private', 'DELETE', {
			orderID : orderId
		});	
	}

	CancelPendingOrders(Currency, BaseCurrency) {
		return this.GetOrders(Currency, BaseCurrency).then(orders=>{
			console.log('cancelling', orders.length, 'orders');
			return Promise.all(orders.map( o => {
				return this.CancelOrder(o.orderID);
			})).then( results => {
				console.log(results);
				return true;
			});
		})
	}

	CloseAllPositions(Currency, BaseCurrency) {
		return this.GetPosition(Currency, BaseCurrency).then(pos=>{
			console.log('closing', pos.length , 'positions');
			return Promise.all( pos.map( p=>{
				if( p.Amount > 0 && p.Type === 'Long' ){
					return this.Trade('Short',0,p.Amount,'XBT', 'USD');
				}
				if( p.Amount > 0 && p.Type === 'Short' ){
					return this.Trade('Long',0,p.Amount,'XBT', 'USD');
				}
				return true;
			})).then(results=>{
				console.log(results);
				return true;
			});
		});
	}

	Trade(direction, price, amount, Currency, BaseCurrency) {
		console.log(direction, price, amount, Currency, BaseCurrency);
		let side = (direction === 'Short' || direction === 'CloseLong') ? 'Sell' : 'Buy';
		let symbol = ( Currency && BaseCurrency) ? Currency + BaseCurrency : this.symbol;
		let params = {
			symbol: symbol,
			side : side,
			ordType: price > 0 ? 'Limit' : 'Market',
			price,
			orderQty: amount
		};
		if ( !price > 0) {
			delete(params.price);
		}

		return this.fetch('order', 'private', 'POST', params)
		.then(o=>{
			if( o && o.orderID ){
				return o.orderID;		
			}else{
				console.log(o);
				throw new Error(o);
			}
			
		});
	}

	GetKline(contract_type, type, size) {
		
	}

	GetDepth(contract_type, size, merge) {
		let params = ['symbol=' + this.symbol];
		if (!size) size = 30;
		if (size) params.push('size=' + size);
		if (merge) params.push('merge=' + merge);
		params.push('contract_type=' + (contract_type||this.options.ContractType));

		return this.fetch('future_depth.do?' + params.join('&'), null, 'GET').then(data => {

			if (!data || !data.bids || !data.asks) throw new Error('get okcoin ' + this.symbol + ' depth error ' + JSON.stringify(data));

			let asks = [];
			let bids = [];

			for (let i = 0;i < data.bids.length;i++) {
				bids.push({
					Price: data.bids[i][0] * 1,
					Amount: data.bids[i][1] * 1
				});
			}

			for (let i = 0;i < data.asks.length;i++) {
				asks.push({
					Price: data.asks[i][0] * 1,
					Amount: data.asks[i][1] * 1
				});
			}

			bids.sort((a, b) => {
				let diff = a.Price - b.Price;
				return diff > 0 ? -1 : diff == 0 ? 0 : 1;
			});

			asks.sort((a, b) => {
				let diff = a.Price - b.Price;
				return diff > 0 ? 1 : diff == 0 ? 0 : -1;
			});

			return Promise.resolve({
				Asks: R.sort( R.descend( R.prop('Price') ), asks),
				Bids: R.sort( R.descend( R.prop('Price') ), bids)
			});
		});
	}
}

module.exports = Exchange;
