const { ok } = require('assert');
const wait = require('delay');
const RateLimiter = require('../lib/rate-limit');
const N = require('precise-number');

class EXCHANGE {
	
	constructor(options) {
		if (!options) options = {};
		ok(options.Currency, 'no Currency in options');
		ok(options.BaseCurrency, 'no BaseCurrency in options');
		ok(options.Name, 'no Name in options');
		ok(options.Fees, 'no Fees in options');
		ok(options.Key, 'no Key in options');
		ok(options.Secret, 'no Secret in options');

		options = Object.assign({
			RateLimit: 10,
			RateLimitInterval: 1000,
			WSTimeout: 10000,
			ThrowWhenRateLimited: false
		}, options);

		if (!options.rateLimiter && options.RateLimit && options.RateLimitInterval) {
			options.rateLimiter = new RateLimiter(options.RateLimitInterval, options.RateLimit, options.ThrowWhenRateLimited);
		}

		this.options = options;
		this.wsReady = false;
		this.approxPrice = null;
	}

	/**
	 * wait websocket ready
	 */
	async waitUntilWSReady() {
		if (!this.options.isWS) return true;
		let startTime = Date.now();
		while (!this.wsReady) {
			await wait(200);
			if (Date.now() - startTime > this.options.WSTimeout) {
				throw new Error(this.GetName() + ' websocket connection timeout');
			}
		}
		return true;
	}

	async Subscribe(Currency, BaseCurrency, type) {
		if (!this.options.isWS) throw new Error('is not websocket mode');
		if (['Depth', 'PublicTrades', 'Ticker'].indexOf(type) === -1) {
			throw new Error('unkown subscription type: ' + type);
		}

		if (type === 'Depth' && !this.options.onDepth) throw new Error('no onDepth callback');
		if (type === 'Ticker' && !this.options.onTicker) throw new Error('no onTicker callback');
		if (type === 'PublicTrades' && !this.options.onPublicTrades) throw new Error('no onPublicTrades callback');

		try {
			await this.waitUntilWSReady();
			await this.ws.addSubscription(Currency, BaseCurrency, type);
		} catch (err) {
			console.error(this.GetName() + ` Subscribe got error:`, err);
			throw err;
		}
	}

	async Init() {
		console.log(this.GetName(), 'does not have Init()');
	}

	/**
	 * set maker and taker fees
	 */
	SetFee(fee) {
		ok(fee, 'no fee');
		ok(fee.Maker);
		ok(fee.Taker);
		this.options.Fees = fee;
	}

	/**
	 * get maker and taker fees
	 */
	GetFee() {
		return this.options.Fees;
	}

	/**
	 * get exchange name
	 */
	GetName() {
		return this.options.Name || 'Unkown Exchange';
	}

	GetMin(price = null) {
		let calculatedAmount = 0;
		if (!price) price = this.approxPrice;
		if (price && this.options.MinTradeValue) {
			calculatedAmount = (this.options.MinTradeValue * 1.2) / price;
		}
		let minAmount = (this.options.MinTradeStocks || this.options.MinTradeAmount) || 0.01;
		return Math.max(minAmount, calculatedAmount);
	}

	SetMin(min) {
		this.options.MinTradeAmount = min;
	}

	async GetMarkets() {
		throw new Error(this.options.Name + ' GetMarkets() method is not implemented');
	}

	async GetAccountsMap(...args) {
		return (await this.GetAccounts(...args)).reduce((map, a) => {
			map[a.Currency] = a;
			return map;
		}, {});
	}

	async GetTickersMap() {
		return (await this.GetTickers()).reduce((map, t) => {
			map[t.Currency + '_' + t.BaseCurrency] = t;
			return map;
		}, {});
	}

	async GetMarketsMap() {
		return (await this.GetMarkets()).reduce((map, t) => {
			map[t.Currency + '_' + t.BaseCurrency] = t;
			return map;
		}, {});
	}

	async GetMarket(Currency, BaseCurrency) {
		if (!this._markets_data) {
			this._markets_data = await this.GetMarketsMap();
		}
		if (!Currency) Currency = this.options.Currency;
		if (!BaseCurrency) BaseCurrency = this.options.BaseCurrency;
		let symbol = Currency + '_' + BaseCurrency;
		if (!this._markets_data[symbol]) throw new Error(`Market ${symbol} not found!`);
		return this._markets_data[symbol];
	}

	async GetFixedRecords(...args) {
		let arr = await this.GetRecords(...args);
		let total = arr.length;
		let timeGap = (args[0] || 1) * 60000;
		let cache = {};
		arr.map(k => {
			cache[k.Time] = k;
		});
		let startTime = arr[0].Time;
		let endTime = arr[arr.length - 1].Time;
		let time = startTime;
		let re = [arr[0]];
		while (time < endTime) {
			time += timeGap;
			let k = cache[time];
			if (!k) {
				k = Object.assign({}, re[re.length - 1]);
				k.Time = time;
			}
			re.push(k);
		}

		return re.slice(-total);
	}

	Buy(...args) {
		return this.Trade('Buy', ...args);	
	}

	Sell(...args) {
		return this.Trade('Sell', ...args);	
	}

	Trade(...args) {
		this._trade_filter(...args);
		return this.rest.Trade(...args);
	}

	_trade_filter(type, price, amount) {
		if (this.options.StockDecimals) amount = N(amount).floor(this.options.StockDecimals);
		if (this.options.Decimals) price = N(price).round(this.options.Decimals);
		this.approxPrice = price;
		if (amount <= 0) throw new Error('trade amount should greater than 0');
		if (this.options.MinTradeValue && amount * price < this.options.MinTradeValue) {
			throw new Error('trade value should greater than ' + this.options.MinTradeValue + `, ${price} * ${amount} = ${N(price).multiply(amount)}`);
		}
	}

	GetTicker(...args) {
		return this.rest.GetTicker(...args);
	}

	GetDepth(...args) {
		return this.rest.GetDepth(...args);
	}

	GetRecords(...args) {
		return this.rest.GetRecords(...args);
	}

	CancelOrder(...args) {
		return this.rest.CancelOrder(...args);
	}

	GetMarkets(...args) {
		return this.rest.GetMarkets(...args);
	}

	GetPublicTrades(...args) {
		return this.rest.GetPublicTrades(...args);
	}

	GetTickers() {
		return this.rest.GetTickers();
	}

	GetAccount(...args) {
		return this.rest.GetAccount(...args);
	}

	GetAccounts(...args) {
		return this.rest.GetAccounts(...args);
	}

	CancelAndGetOrder(...args) {
		return this.CancelOrder(...args).then(() => {
			return this.GetOrder(...args);
		});
	}

	GetClientId(ClientId) {
		if (ClientId) return ClientId;
		if (this.options.FixedClientId) return `${Date.now()}${String(this.options.FixedClientId).padStart(6, '0')}`;
		return null;
	}
}


module.exports = EXCHANGE;
