const EXCHANGE_REST = require('./rest.js');
const EXCHANGE_WS = require('./ws.js');
const N = require('precise-number');
const { ok } = require('assert');
const fetch = require('node-fetch');
const EXCHANGE = require('../exchange.js');

class HBDM extends EXCHANGE {
	constructor(options) {
		options = Object.assign({
			Name: 'HBDM',
			Fees: {
				Maker: 0.0001,
				Taker: 0.0003
			},
			RateLimit: 10,
			Decimals: 8,
			StockDecimals: 3,
			MinTradeAmount: 1,
			BaseCurrency: 'USD',
			ContractType: 'this_week',
			Currency: 'BTC',
			ContractMode: 'All',
			MarginLevel: 10
		}, options);
		super(options);

		this.options.BaseCurrency = 'USD';
		if (['All', 'Seperate'].indexOf(options.ContractMode) === -1) throw new Error('options.ContractMode should either be All or Seperate');
		if (options.MarginLevel !== 10 && options.MarginLevel !== 20) throw new Error('wrong MarginLevel');

		this._check_contract_type(options.ContractType);

		// this.options = options;
		if (options.isWS) {
			this.ws = new EXCHANGE_WS(this.options);
			this.ws.on('connect', () => {
				this.wsReady = true;
			});
			this.ws.on('close', () => {
				this.wsReady = false;
			});
		}
		this.rest = new EXCHANGE_REST(this.options);
	}

	_check_contract_type(contract_type) {
		if (['this_week', 'next_week', 'quarter'].indexOf(contract_type) === -1) {
			throw new Error('unkown contract_type: ' + contract_type);
		}
	}

	_on_trade(cb) {
		return (err, o) => {
			if (err) {
				cb(err, o);
				return;
			}
			cb(false, {
				Id: o.orderId,
				Price: N.parse(o.tradePrice),
				AvgPrice: N.parse(o.averagePrice),
				Amount: N.parse(o.tradeAmount),
				DealAmount: N.parse(o.completedTradeAmount),
				Status: this._order_status(o.status),
				Type: this._order_type(o.tradeType),
				LeftAmount: N.parse(o.unTrade),
				UnitPrice: N.parse(o.tradeUnitPrice),
				SigAmount: N.parse(o.sigTradeAmount),
				SigPrice: N.parse(o.sigTradePrice),
				Time: N.parse(o.createdDate)
			});
		};
	}

	_on_account_change(cb) {
		return (err, data) => {
			if (err) {
				cb(err, data);
				return;
			}
			let currency = this.options.Currency.toLowerCase();
			cb(false, {
				Balance: N.parse(data.info.free.usd),
				FrozenBalance: N.parse(data.info.freezed.usd),
				Stocks: N.parse(data.info.free[currency]),
				FrozenStocks: N.parse(data.info.freezed[currency])
			});
		};
	}

	async Subscribe(Currency, BaseCurrency, type, contractType) {
		if (!this.options.isWS) throw new Error('is not websocket mode');
		if (['Depth', 'PublicTrades', 'Ticker'].indexOf(type) === -1) {
			throw new Error('unkown subscription type: ' + type);
		}

		if (contractType) this._check_contract_type(contractType);

		if (type === 'Depth' && !this.options.onDepth) throw new Error('no onDepth callback');
		if (type === 'Ticker' && !this.options.onTicker) throw new Error('no onTicker callback');
		if (type === 'PublicTrades' && !this.options.onPublicTrades) throw new Error('no onPublicTrades callback');

		try {
			await this.waitUntilWSReady();
			await this.ws.addSubscription(Currency, BaseCurrency, type, contractType);
		} catch (err) {
			console.error(this.GetName() + ` Subscribe got error:`, err);
			throw err;
		}
	}

	async GetPublicTrades(Currency, BaseCurrency, contract_type, size = 200) {
		return this.rest.GetPublicTrades(Currency, BaseCurrency, contract_type, size);
	}

	API(url, params, method) {
		if (!method) method = 'POST';
		return this.rest.fetch(url, params, method);
	}

	//设置杠杆倍率
	SetMarginLevel(level) {
		this.options.MarginLevel = level;
	}

	getHandler() {
		return this.rest;
	}

	GetAccount(...args) {
		return this.rest.GetAccount(...args);
	}

	GetPosition(Currency, BaseCurrency, contract_type) {
		return this.rest.GetPosition(Currency, BaseCurrency, contract_type);
	}

	async GetPositionMap(Currency, BaseCurrency, contract_type) {
		let positions = await this.GetPosition(Currency, BaseCurrency, contract_type);
		let re = {
			Long: {
				MarginLevel: 0,
				Amount: 0,
				FrozenAmount: 0,
				Price: 0,
				Profit: 0,
				Type: 'Long',
				ContractType: contract_type || this.options.ContractType,
				Currency: Currency || this.options.Currency,
				BaseCurrency: BaseCurrency || this.options.BaseCurrency
			},
			Short: {
				MarginLevel: 0,
				Amount: 0,
				FrozenAmount: 0,
				Price: 0,
				Profit: 0,
				Type: 'Short',
				ContractType: contract_type || this.options.ContractType,
				Currency: Currency || this.options.Currency,
				BaseCurrency: BaseCurrency || this.options.BaseCurrency
			}
		};
		positions.map(p => {
			if (p.Type === 'Long') {
				re.Long = p;
			} else if (p.Type === 'Short') {
				re.Short = p;
			}
		});
		return re;
	}

	async GetNetPosition(Currency, BaseCurrency, contract_type) {
		let positions = await this.GetPositionMap(Currency, BaseCurrency, contract_type);
		let Amount = positions.Long.Amount - positions.Short.Amount;
		return {
			Amount,
			ContractType: contract_type || this.options.ContractType,
			Price: Amount > 0 ? positions.Long.Price : positions.Short.Price,
			MarginLevel: Amount > 0 ? positions.Long.MarginLevel : positions.Short.MarginLevel,
			Currency: Currency || this.options.Currency,
			BaseCurrency: BaseCurrency || this.options.BaseCurrency,
			Profit: positions.Long.Profit - positions.Short.Profit
		};
	}

	GetTicker(...args) {
		return this.rest.GetTicker(...args);
	}

	GetOrder(orderId, Currency, BaseCurrency, contract_type) {
		if (!orderId || orderId === 'undefined') throw new Error('GetOrder orderId error:' + orderId);
		return this.rest.GetOrder(orderId, Currency, BaseCurrency, contract_type).then(data => {
			let orders = data && data.data ? data.data : [];
			orders = orders.map(o => {
				return this._transform_order(o);
			});
			if (orders && orders.length > 0) return orders[0];
			throw new Error('can not get order info, id = ' + orderId);
		});
	}

	GetOrders(Currency) {
		return this.rest.GetOrders(Currency).then(orders => orders.map(this._transform_order.bind(this)));
	}

	_transform_order(o) {
		return {
			Id: o.order_id,
			Price: N.parse(o.price),
			Amount: N.parse(o.volume),
			DealAmount: N.parse(o.trade_volume),
			Fee: N.parse(o.fee),
			Status: this._order_status(o.status),
			Type: this._transform_type(o.direction, o.offset),
			AvgPrice: N.parse(o.trade_avg_price),
			MarginLevel: N.parse(o.lever_rate),
			Currency: o.symbol,
			BaseCurrency: 'USD',
			ContractType: o.contract_type,
			Time: o.created_at,
			ClientOrderId: o.client_order_id,
			Info: o
		};
	}

	_transform_type(direction, offset) {
		if (direction === 'buy') {
			return offset === 'open' ? 'Long' : 'CloseShort';
		} else {
			return offset === 'open' ? 'Short' : 'CloseLong';
		}
	}

	Trade(direction, price, amount, Currency, BaseCurrency, contract_type, options) {
		if (!contract_type) contract_type = this.options.ContractType;
		this._check_contract_type(contract_type);
		if (['Long', 'Short', 'CloseLong', 'CloseShort'].indexOf(direction) === -1) {
			throw new Error('unknown direction: ' + direction + ', direction should be one of these values: Long Short CloseLong CloseShort');
		}
		ok( amount > 0, 'amount should greater than 0');

		if (typeof Currency === 'object') {
			options = Currency;
			Currency = undefined;
		}
		if (typeof BaseCurrency === 'object') {
			options = BaseCurrency;
			BaseCurrency = undefined;
		}
		if (typeof contract_type === 'object') {
			options = contract_type;
			contract_type = undefined;
		}
		if (!options) options = {};
		options.ClientId = this.GetClientId(options.ClientId);

		console.log(this.GetName(contract_type), direction, price, amount, options);
		return this.getHandler().Trade(direction, price, amount, Currency, BaseCurrency, contract_type, options).then( r => {
			if (r && r.status === 'ok' && r.data && r.data.order_id) {
				return r.data.order_id;
			} else {
				console.log(r);
				throw new Error(r);
			}
		});
	}

	CancelOrder(orderId, Currency) {
		return this.getHandler().CancelOrder(orderId, Currency);
	}

	CancelPendingOrders(...args) {
		console.log('cancelling pending orders...');
		return this.rest.CancelPendingOrders(...args);
	}

	GetDepth(...args) {
		return this.rest.GetDepth(...args);
	}

	async GetAvgPriceShift(contract_type) {
		let [k1, k2] = await Promise.all([
			this.GetIndexKline(),
			this.GetKline(contract_type)
		]);
		let k1hash = {}, k2hash = {};
		k1 = k1.map(k => {
			let m = Math.floor(k.Time / 60000);
			k1hash[m] = k.Close;
			k.m = m;
			return k;
		});
		k2 = k2.map(k => {
			let m = Math.floor(k.Time / 60000);
			k2hash[m] = k.Close;
			k.m = m;
			return k;
		});
		let startMinute = Math.max( Math.min.apply(null, k1.map(k => k.m)), Math.min.apply(null, k2.map(k => k.m)));
		let endMinute = Math.min( Math.max.apply(null, k1.map(k => k.m)), Math.max.apply(null, k2.map(k => k.m)));

		let totalDiff = 0, t = 0;
		for (let i = startMinute; i <= endMinute; i++) {
			if (k1hash[i] && k2hash[i]) {
				t ++;
				totalDiff += k1hash[i] - k2hash[i];
			}
		}
		if (t === 0) return 0;
		return totalDiff / t;
	}

	GetKline(contract_type, n) {
		if (!n) n = 100;
		if (!contract_type) contract_type = this.options.ContractType;
		let url = `https://www.okex.com/v2/futures/pc/market/klineData.do?symbol=f_usd_${this.options.Currency.toLowerCase()}&type=1min&limit=${n}&coinVol=1&contractType=${contract_type}`;
		return fetch(url).then(res => res.json()).then(data => {
			if (data && data.data && data.data.length > 0) {
				return data.data.map(h => {
					let [Time, Open, High, Low, Close, Volume] = h;
					return {Open, High, Low, Close, Volume, Time};
				});
			}
		});
	}

	GetMarkets() {
		return this.rest.GetMarkets();
	}

	GetTickers() {
		return this.rest.GetTickers();
	}

	GetAccounts() {
		return this.rest.GetAccounts();
	}

	_order_status( status ) {
		//(3未成交 4部分成交 5部分成交已撤单 6全部成交 7已撤单)
		status = status.toString();
		let arr = {
			'5':'Cancelled',
			'1': 'Pending',
			'2': 'Pending',
			'3': 'Pending',
			'4': 'Partial',
			'6': 'Closed',
			'7': 'Cancelled',
			'11': 'Pending'
		};
		return arr[status + ''];
	}

	_order_type( type ) {
		//订单类型 1：开多 2：开空 3：平多 4： 平空
		type = type.toString();
		let arr = {
			'1': 'Long',
			'2': 'Short',
			'3': 'CloseLong',
			'4': 'CloseShort'
		};
		return arr[`${type}`];
	}

	GetRecords(minutes, Currency, BaseCurrency, ContractType) {
		return this.rest.GetRecords(minutes, Currency, BaseCurrency, ContractType);
	}

}


module.exports = HBDM;
