const fetch = require('node-fetch');
const N = require('precise-number');
const R = require('ramda');
const debug = require('debug')('exchange:hbdm:rest');
const agent = require('../../lib/agent');
const ExError = require('../../lib/error');
const ErrorCode = require('../../lib/error-code');
const crypto = require('crypto');

class HBDM {

	constructor(options) {
		this.key = options.Key;
		this.secret = options.Secret;
		this.options = options;
		if (!this.options.rateLimiter) throw 'No rateLimiter in options';
	}

	async fetch(httpMethod, method, params, body, isPublic = false) {
		await this.options.rateLimiter.wait();
		if (!params) params = {};

		if (!isPublic) {
			params.AccessKeyId = this.options.Key;
			params.SignatureMethod = 'HmacSHA256';
			params.SignatureVersion = '2';
			params.Timestamp = (new Date()).toISOString().substr(0, 19);
		}

		let sortedParams = R.compose(
			R.join('&'),
			R.sort(R.ascend(R.identity)),
			R.map(a => a[0] + '=' + encodeURIComponent(a[1])),
			R.toPairs
		)(params);

		let signString = [httpMethod, 'api.hbdm.com', method, sortedParams].join('\n');

		if (!isPublic) {
			let hash = crypto.createHmac('sha256', Buffer.from(this.options.Secret, 'utf8')).update(signString).digest('base64');
			sortedParams += '&Signature=' + encodeURIComponent(hash);
		}

		let url = 'https://api.hbdm.com' + method + '?' + sortedParams;
		debug('<<<', httpMethod, url, JSON.stringify(body || 'nobody'));
		return fetch(url, {
			method: httpMethod,
			timeout: httpMethod === 'GET' ? 5000 : 10000,
			body: httpMethod === 'POST' ? (body !== undefined ? JSON.stringify(body) : undefined) : undefined,
			headers: {
				"Content-Type": "application/json",
				"User-Agent": "Mozilla/5.0 (Windows NT 6.1; WOW64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/39.0.2171.71 Safari/537.36"
			},
			agent: agent.https
		}).then(async res => {
			let s = await res.text();
			let status = res.status;

			if (status === 429) throw new ExError(ErrorCode.REQUEST_TOO_FAST, '429 request too fast');
			if (status !== 200) throw new ExError(ErrorCode.BAD_RESPONSE_STATUS, `Huobi response bad status(${status})`, s);

			debug('>>>', s);
			let data = null;
			try {
				s = s.replace(/(\:\s*)(\d{14,})([^\"])/g, '$1"$2"$3');
				data = JSON.parse(s);
			} catch (err) {
				throw new ExError(ErrorCode.INVALID_JSON, 'HuobiProRest can not parse json', err);
			}

			if (data && data.status === 'error') {

				let err = new Error(data.err_msg || data.err_code || JSON.stringify(data));
				err.code = data.err_code;
				err.url = url;
				err.body = body;
				err.error_code = data.err_code;
				throw err;
			}

			if (!data || data.status !== 'ok') {
				this.error('api response error:', data);
			} else {
				return data;
			}
		}).catch(err => {
			if (err.type === 'request-timeout') {
				throw new ExError(ErrorCode.REQUEST_TIMEOUT, `huobi rest request (${method}) timeout`, err);
			}
			throw err;
		});
	}

	_getSymbol(Currency, BaseCurrency, ContractType) {
		if (!Currency) Currency = this.options.Currency;
		if (!ContractType) ContractType = this.options.ContractType;
		if (BaseCurrency && BaseCurrency !== 'USD') throw new Error('BaseCurrency should be USD');
		let names = {
			this_week: 'CW',
			next_week: 'NW',
			quarter: 'CQ'
		};
		return `${Currency}_${names[ContractType]}`.toUpperCase();
	}

	_symbolNameToType(name) {
		let types = {
			CW: 'this_week',
			NW: 'next_week',
			CQ: 'quarter'
		};
		return types[name];
	}

	_parseSymbol(Currency, BaseCurrency, ContractType) {
		if (!Currency) Currency = this.options.Currency;
		if (!BaseCurrency) BaseCurrency = this.options.BaseCurrency;
		if (!ContractType) ContractType = this.options.ContractType;
		return { Currency, BaseCurrency, ContractType };
	}

	_parseSymbolString(symbol) {
		if (!symbol) symbol = this.symbol;
		let ms = symbol.match(/(\w+)\_(CW|NW|CQ)/);
		if (ms && ms[1] && ms[2]) {
			let ContractType = this._symbolNameToType(ms[2]);
			let Currency = String(ms[1]).toUpperCase();
			let BaseCurrency = 'USD';
			return { Currency, BaseCurrency, ContractType };
		}
		throw new Error('can not parse symbol: ' + symbol);
	}

	GetTicker(Currency, BaseCurrency, contract_type) {
		return this.fetch('GET', `/market/detail/merged`, {
			symbol: this._getSymbol(Currency, BaseCurrency, contract_type)
		}, '', true).then(data => {
			return Promise.resolve({
				High: N.parse(data.tick.high),
				Low: N.parse(data.tick.low),
				Buy: N.parse(data.tick.bid[0]),
				Sell: N.parse(data.tick.ask[0]),
				Last: N.parse(data.tick.bid[0]),
				Volume: N.parse(data.tick.vol),
				Time: N.parse(data.tick.ts),
				...this._parseSymbolString(data.ch),
				Info: data
			});
		});
	}

	GetPublicTrades(Currency, BaseCurrency, contract_type, size = 200) {
		return this.fetch('GET', `/market/history/trade`, {
			symbol: this._getSymbol(Currency, BaseCurrency, contract_type),
			size
		}, null, true).then(data => {
			let re = [];
			let info = this._parseSymbolString(data.ch);
			if (data && data.data) {
				data.data.map(trade => {
					if (trade && trade.data) {
						trade.data.map(t => {
							re.push({
								Amount: N.parse(t.amount),
								Id: t.id,
								Price: N.parse(t.price),
								Time: N.parse(t.ts),
								Type: t.direction === 'sell' ? 'Sell' : 'Buy',
								...info
							});
						});
					}
				});
			}
			return re;
		});
	}

	GetAccount(Currency, BaseCurrency) {
		if (!Currency) Currency = this.options.Currency;
		if (!BaseCurrency) BaseCurrency = this.options.BaseCurrency;
		return this.fetch('POST', '/api/v1/contract_account_info', null, {
			symbol: Currency
		}).then(data => {
			let info = data.data[0];
			let frozen = N(info.margin_frozen).add(info.margin_position) * 1;
			let re = {
				Balance: 0,
				FrozenBalance: 0,
				Stocks: N(info.margin_balance).minus(frozen) * 1,
				FrozenStocks: frozen,
				Value: N.parse(info.margin_balance),
				Risk: N.parse(info.risk_rate),
				Currency,
				BaseCurrency,
				Info: data
			};
			return re;
		});
	}

	GetAccounts() {
		return this.fetch('POST', '/api/v1/contract_account_info').then(data => {
			if (data && data.data && data.data.length > 0) {
				return data.data.map(a => {
					let frozen = N(a.margin_frozen).add(a.margin_position) * 1;
					return {
						Currency: a.symbol,
						Free: N(a.margin_balance).minus(frozen) * 1,
						Frozen: frozen,
						Value: N.parse(a.margin_balance),
						Risk: N.parse(a.risk_rate),
						Info: a
					};
				});
			}
			throw new Error('GetAccounts returns bad data: ' + JSON.stringify(data));
		});
	}

	GetPosition(Currency, BaseCurrency, contract_type) {
		if (!Currency) Currency = this.options.Currency;
		if (!BaseCurrency) BaseCurrency = this.options.BaseCurrency;
		if (!contract_type) contract_type = this.options.ContractType;
		return this.fetch('POST', '/api/v1/contract_position_info', null, {
			symbol: Currency
		}).then(data => {
			let re = [];
			if (data && data.data && Array.isArray(data.data)) {
				data.data.map(h => {
					if (h.contract_type === contract_type || contract_type === 'ALL') {
						if (h.direction === 'buy') {
							re.push({
								MarginLevel: h.lever_rate,
								Amount: N.parse(h.available),
								FrozenAmount: N.parse(h.frozen),
								Price: N.parse(h.cost_open),
								Profit: N.parse(h.profit_unreal),
								Type: 'Long',
								ContractType: h.contract_type,
								Currency,
								BaseCurrency: 'USD',
								Info: h
							});
						} else if (h.direction === 'sell') {
							re.push({
								MarginLevel: h.lever_rate,
								Amount: N.parse(h.available),
								FrozenAmount: N.parse(h.frozen),
								Price: N.parse(h.cost_open),
								Profit: N.parse(h.profit_unreal),
								Type: 'Short',
								ContractType: h.contract_type,
								Currency,
								BaseCurrency: 'USD',
								Info: h
							});
						}
					}
				});
			}
			return re;
		});
	}

	CancelPendingOrders(Currency, BaseCurrency) {
		if (!Currency) Currency = this.options.Currency;
		return this.fetch('POST', '/api/v1/contract_cancelall', null, {
			symbol: Currency
		}).then(data => {
			if (data && data.data) {
				if (data.data.errors && data.data.errors.length > 0) {
					throw new Error('CancelPendingOrders has error:' + JSON.stringify(data.data.errors));
				} else {
					return data.data.successes;
				}
			}
			throw new Error('CancelPendingOrders got wrong response:' + JSON.stringify(data));
		}).catch(err => {
			//{"status":"error","err_code":1051,"err_msg":"No orders to cancel.","ts":1544619550743}
			if (err && err.error_code === 1051) return [];
			throw err;
		});
	}

	GetHoldAmount(Currency, BaseCurrency, contract_type) {
		return this.fetch(`future_hold_amount.do?symbol=${this._getSymbol(Currency, BaseCurrency)}&contract_type=${contract_type || this.options.ContractType}`, null, 'GET');
	}

	GetOrder(orderId, Currency) {
		let params = {
			symbol: Currency || this.options.Currency,
			order_id: String(orderId)
		};
		return this.fetch('POST', '/api/v1/contract_order_info', null, params);
	}

	GetOrders(Currency) {
		let params = {
			symbol: Currency || this.options.Currency
		};
		return this.fetch('POST', '/api/v1/contract_openorders', null, params).then(d => d.data.orders);
	}

	CancelOrder(orderId, Currency) {
		return this.fetch('POST', '/api/v1/contract_cancel', null, {
			symbol: Currency || this.options.Currency,
			order_id: orderId
		}).then(data => {
			if (data && data.status === 'ok' && String(data.data.successes) === String(orderId)) {
				return true;
			} else {
				if (data && data.data && data.data.errors && data.data.errors.length > 0) {
					let errors = data.data.errors.filter(e => e.err_code !== 1071);
					if (data.data.errors.length > 0 && errors.length === 0) return true;
				}
				console.error('cancel order got error:', JSON.stringify(data.data, null, '\t'));
			}
		});
	}

	Trade(direction, price, amount, Currency, BaseCurrency, contract_type, options = {}) {
		let params = {
			symbol: Currency || this.options.Currency,
			contract_type: contract_type || this.options.ContractType,
			price,
			volume: amount,
			direction: ['Long', 'CloseShort'].indexOf(direction) > -1 ? 'buy' : 'sell',
			offset: ['Long', 'Short'].indexOf(direction) > -1 ? 'open' : 'close',
			lever_rate: this.options.MarginLevel,
			order_price_type: this.getOrderType(options.Type)
		};

		if (options.ClientId) params.client_order_id = options.ClientId;

		return this.fetch('POST', '/api/v1/contract_order', null, params);
	}

	getOrderType(Type) {
		//订单报价类型 "limit":限价 "opponent":对手价 "post_only":只做maker单,post only下单只受用户持仓数量限制,optimal_5：最优5档、optimal_10：最优10档、optimal_20：最优20档，ioc:IOC订单，fok：FOK订单
		const types = {
			GTC: 'limit',
			PO: 'post_only',
			IOC: 'ioc',
			FOK: 'fok'
		};
		if (!Type) return this.options.MakerMode ? types.PO : types.GTC;
		return types[Type] || Type;
	}

	GetDepth(Currency, BaseCurrency, ContractType, type) {
		let symbol = this._getSymbol(Currency, BaseCurrency, ContractType);
		if (!type) type = 'step0';

		return this.fetch('GET', '/market/depth', {
			symbol,
			type
		}, null, true).then(data => {

			let tick = data.tick;
			if (!data || !tick || !tick.bids || !tick.asks) throw new Error('get ' + symbol + ' depth error ' + JSON.stringify(data));

			let asks = [];
			let bids = [];

			for (let i = 0;i < tick.bids.length;i++) {
				bids.push({
					Price: tick.bids[i][0] * 1,
					Amount: tick.bids[i][1] * 1
				});
			}

			for (let i = 0;i < tick.asks.length;i++) {
				asks.push({
					Price: tick.asks[i][0] * 1,
					Amount: tick.asks[i][1] * 1
				});
			}

			bids.sort((a, b) => {
				let diff = a.Price - b.Price;
				return diff > 0 ? -1 : diff == 0 ? 0 : 1;
			});

			asks.sort((a, b) => {
				let diff = a.Price - b.Price;
				return diff > 0 ? 1 : diff == 0 ? 0 : -1;
			});

			return Promise.resolve({
				Asks: R.sort( R.ascend( R.prop('Price') ), asks),
				Bids: R.sort( R.descend( R.prop('Price') ), bids),
				...this._parseSymbolString(data.ch)
			});
		});
	}

	async GetTickers() {
		let options = {
			method: 'GET',
			timeout: 5000,
			headers: {},
			agent: agent.https
		};

		let types = [
			'',
			'this_week',
			'next_week',
			'',
			'quarter'
		];

		let data = await fetch('https://www.okex.com/v2/futures/pc/market/marketOverview.do?symbol=f_usd_all', options).then(res => res.json());
		if (data && data.ticker && Array.isArray(data.ticker)) {
			return data.ticker.map(t => {
				return {
					High: N.parse(t.high),
					Low: N.parse(t.low),
					Buy: N.parse(t.buy),
					Sell: N.parse(t.sell),
					Last: N.parse(t.last),
					Volume: N.parse(t.volume),
					Time: Date.now(),
					Currency: String(t.symbolName).toUpperCase(),
					BaseCurrency: 'USD',
					ContractType: types[t.contractType],
					Info: t
				};
			});
		} else {
			throw new Error('can not get okex tickers: ' + JSON.stringify(data));
		}
	}

	async GetMarkets() {
		let data = await this.fetch('GET', '/api/v1/contract_contract_info', null, null, true);

		return data.data.map(m => {
			let d = m.delivery_date;
			d = `${d.substr(0, 4)}-${d.substr(4, 2)}-${d.substr(6, 2)}T08:00:00.000Z`;
			d = new Date(d);
			return {
				Currency: m.symbol,
				BaseCurrency: 'USD',
				ContractType: m.contract_type,
				ContractId: m.contract_code,
				DeliveryTime: d.getTime(),
				Decimals: m.price_tick ? Math.round(Math.log10(1 / m.price_tick)) : 0,
				StockDecimals: 0,
				MinTradeAmount: 1,
				UnitValue: Math.round(m.contract_size),
				Info: m
			};
		});
	}

	async GetRecords(minutes = 1, Currency, BaseCurrency, ContractType) {
		let period = '';
		if (minutes <= 60) period = minutes + 'min';
		if (minutes === 240) period = '4h';
		if (minutes === 1440) period = '1d';
		if (!period) throw new Error('wrong records period: ' + minutes);
		let data = await this.fetch('GET', `/market/history/kline`, {
			period,
			size: 200,
			symbol: this._getSymbol(Currency, BaseCurrency, ContractType)
		}, null, true);
		return data.data.map(k => {
			return {
				Open: N.parse(k.open),
				High: N.parse(k.high),
				Low: N.parse(k.low),
				Close: N.parse(k.close),
				Volume: N.parse(k.vol),
				CurrencyVolume: N.parse(k.amount),
				Time: k.id * 1000
			};
		}).sort((a, b) => {
			return a.Time - b.Time;
		});
	}
}


module.exports = HBDM;
