const N = require('precise-number');
const WebSocket = require('../../lib/auto-reconnect-ws.js');
const Events = require('events');
const zlib = require('zlib');


class HBDM_WS extends Events {
	constructor(options) {
		super();
		this.options = options;

		this.lastPingTime = Date.now();
		this.lastMsgTime = 0;

		this.DepthMode = this.options.WSDepthMode || 'step6';

		this.ws = new WebSocket('wss://www.hbdm.com/ws', {
			pingInterval: 5000
		});

		//remember subscription commands
		this.subscriptionCommands = [];

		if (this.options.onDepth) {
			this.subscriptionCommands.push(JSON.stringify({
				sub: `market.${this._getSymbol()}.depth.${this.DepthMode}`
			}));
		}

		if (this.options.onPublicTrades) {
			this.subscriptionCommands.push(JSON.stringify({
				sub: `market.${this._getSymbol()}.trade.detail`
			}));
		}

		if (this.options.onTicker) {
			this.subscriptionCommands.push(JSON.stringify({
				sub: `market.${this._getSymbol()}.detail`
			}));
		}

		this.wsReady = false;
		this.ws.on('open', () => {
			this.subscriptionCommands.map(cmd => {
				this.ws.send(cmd);
			});
			this.emit('connect');
		});

		this.ws.on('message', (data) => {
			this.lastMsgTime = Date.now();
			this.wsReady = true;
			this.onMessage(data);
		});

		this.ws.on('pong', t => {
			this.wsReady = true;
			if (this.options.onPong) {
				this.options.onPong(t);
			}
		});

		this.ws.on('close', () => {
			console.log('ws close');
			this.emit('close');
			this.wsReady = false;
		});

		this.ws.on('error', () => {
			console.log('ws error');
			this.emit('close');
			this.wsReady = false;
		});
	}

	_getSymbol(Currency, BaseCurrency, ContractType) {
		if (!Currency) Currency = this.options.Currency;
		if (!ContractType) ContractType = this.options.ContractType;
		if (BaseCurrency && BaseCurrency !== 'USD') throw new Error('BaseCurrency should be USD');
		let names = {
			this_week: 'CW',
			next_week: 'NW',
			quarter: 'CQ'
		};
		return `${Currency}_${names[ContractType]}`.toUpperCase();
	}

	onMessage(raw)  {
		if (typeof raw !== 'string') {
			raw = zlib.gunzipSync(raw);
		}
		
		let message = null;
		try {
			message = JSON.parse(raw);
		} catch (err) {
			console.error('okex websocket message error: ' + raw);
			return;
		}
		if (!message) return;

		if (message.ping) {
			this.ws.send(JSON.stringify({
				pong: message.ping
			}));
			return;
		}

		if (message['err-code']) {
			let err = new Error(message['err-msg']);
			err.error_code = message['err-code'];
			err.original = message;
			throw err;
		}

		let { ch: channel } = message;
		if (!channel) return;
		if (channel.match(/depth/)) {
			this.onDepth(message);
		} else if (channel.match(/trade/)) {
			this.onPublicTrades(message);
		} else if (channel.match(/market\.[^\.]+\.detail$/)) {
			this.onTicker(message);
		} else {
			console.log('unkown message', raw);
		}
	}

	addSubscription(Currency, BaseCurrency, type, ContractType) {
		if (!Currency) Currency = this.options.Currency;
		if (!BaseCurrency) BaseCurrency = this.options.BaseCurrency;
		if (!ContractType) ContractType = this.options.ContractType;

		let channel = '';
		if (type === 'Depth') {
			channel = `market.${this._getSymbol(Currency, BaseCurrency, ContractType)}.depth.${this.DepthMode}`;
		} else if (type === 'PublicTrades') {
			channel = `market.${this._getSymbol(Currency, BaseCurrency, ContractType)}.trade.detail`;
		}

		if (!channel) throw new Error('unkown subscription type: ' + type);

		let cmd = JSON.stringify({
			sub: channel
		});

		this.ws.send(cmd);
		this.subscriptionCommands.push(cmd);
	}

	onDepth(data) {
		let { ch: channel, tick: { bids, asks, ts: timestamp } } = data;
		if (!asks || !bids || !channel || !timestamp) return;

		asks = asks.map(pair => {
			return {
				Price: pair[0] * 1,
				Amount: pair[1] * 1
			};
		});

		bids = bids.map(pair => {
			return {
				Price: pair[0] * 1,
				Amount: pair[1] * 1
			};
		});

		let depth = {
			Time: N.parse(timestamp),
			Asks: asks, //R.sort( R.ascend( R.prop('Price') ), asks),
			Bids: bids, //R.sort( R.descend( R.prop('Price') ), bids),
			...this._parse_ch(channel)
		};

		this.options.onDepth(depth, this.lastMsgTime);
	}

	_symbolNameToType(name) {
		let types = {
			CW: 'this_week',
			NW: 'next_week',
			CQ: 'quarter'
		};
		return types[name];
	}

	_parse_ch(symbol) {
		if (!symbol) symbol = this.symbol;
		let ms = symbol.match(/(\w+)\_(CW|NW|CQ)/);
		if (ms && ms[1] && ms[2]) {
			let ContractType = this._symbolNameToType(ms[2]);
			let Currency = String(ms[1]).toUpperCase();
			let BaseCurrency = 'USD';
			return { Currency, BaseCurrency, ContractType };
		}
		throw new Error('can not parse symbol: ' + symbol);
	}

	onTicker({ch: channel, tick: data, ts: timestamp}) {
		if (!data || !channel || !timestamp) return;
		let re = {
			Open: N.parse(data.open),
			High: N.parse(data.high),
			Low: N.parse(data.low),
			Close: N.parse(data.close),
			Last: N.parse(data.close),
			Volume: N.parse(data.vol),
			Time: N.parse(timestamp),
			...this._parse_ch(channel)
		};
		if (this.options.onTicker) {
			this.options.onTicker(re);
		}
	}

	onPublicTrades(data) {
		let { ch: channel, tick: { data: trades}, ts: timestamp } = data;
		if (!channel || !trades || trades.length <= 0 || !timestamp) return;

		trades = trades.map(t => ({
			Amount: N.parse(t.amount),
			Price: N.parse(t.price),
			Type: t.direction === 'buy' ? 'Buy' : 'Sell',
			Time: N.parse(t.ts),
			Id: t.id,
			...this._parse_ch(channel)
		}));
		this.options.onPublicTrades(trades);
	}
}


module.exports = HBDM_WS;
