const EXCHANGE_REST = require('./rest.js');
const EXCHANGE_WS = require('./ws.js');
const N = require('precise-number');
const { ok } = require('assert');
const EXCHANGE = require('../exchange.js');

class KuCoin extends EXCHANGE {
	constructor(options) {
		options = Object.assign({
			Name: 'KuCoin',
			Fees: {
				Maker: 0.002,
				Taker: 0.002
			},
			RateLimit: 10,
			MakerMode: false
		}, options);
		super(options);

		this.rest = new EXCHANGE_REST(this.options);
	
		if (options.isWS) {
			this.ws = new EXCHANGE_WS(this.options, this.rest);
			this.ws.on('connect', () => {
				this.wsReady = true;
			});
			this.ws.on('close', () => {
				this.wsReady = false;
			});
		}
	}

	async Subscribe(currency, baseCurrency, type) {
		if (!this.options.isWS) throw new Error('is not websocket mode');
		if (['Depth', 'PublicTrades', 'Ticker'].indexOf(type) === -1) {
			throw new Error('unkown subscription type: ' + type);
		}

		if (type === 'Depth' && !this.options.onDepth) throw new Error('no onDepth callback');
		if (type === 'Ticker' && !this.options.onTicker) throw new Error('no onTicker callback');
		if (type === 'PublicTrades' && !this.options.onPublicTrades) throw new Error('no onPublicTrades callback');

		try {
			await this.waitUntilWSReady();
			await this.ws.addSubscription(currency, baseCurrency, type);
		} catch (err) {
			console.error(this.GetName() + ` Subscribe got error:`, err);
			throw err;
		}
	}

	GetAccount(Currency, BaseCurrency, accountType) {
		Currency = Currency || this.options.Currency;
		BaseCurrency = BaseCurrency || this.options.BaseCurrency;

		return this.GetAccountsMap(accountType).then(data => {
			// console.log('get accounts map', data);
			return {
				Balance: data[BaseCurrency] ? data[BaseCurrency].Free : 0,
				FrozenBalance: data[BaseCurrency] ? data[BaseCurrency].Frozen : 0,
				Stocks: data[Currency] ? data[Currency].Free : 0,
				FrozenStocks: data[Currency] ? data[Currency].Frozen : 0,
				Currency,
				BaseCurrency
			};
		});
	}

	GetAccounts(type = 'trade', currency = '') {
		if (!type) type = 'trade';
		if (type && type !== 'main' && type !== 'trade') throw new Error('type should be main or trade');
		return this.rest.GetAccounts(type, currency);
	}

	GetTicker(Currency, BaseCurrency) {
		return this.rest.GetTicker(Currency, BaseCurrency);
	}

	GetOrder(orderId, Currency, BaseCurrency) {
		return this.rest.GetOrder(orderId, Currency, BaseCurrency).then(o => {
			return {
				Id: o.id,
				Price: N.parse(o.price),
				Amount: N.parse(o.size),
				DealAmount: N.parse(o.dealSize),
				Status: this._order_status(o),
				Type: this._order_type(o.side),
				AvgPrice: N.parse(o.price),
				Time: N.parse(o.createdAt),
				...this.rest._parse_ch(o.symbol),
				Info: o
			};
		});
	}

	async GetOrders(...args) {
		let orders = await this.rest.GetOrders(...args);
		return orders.map(o => {
			return {
				Id: o.id,
				Price: N.parse(o.price),
				Amount: N.parse(o.size),
				DealAmount: N.parse(o.dealSize),
				Status: this._order_status(o),
				Type: this._order_type(o.side),
				AvgPrice: N.parse(o.price),
				Time: N.parse(o.createdAt),
				...this.rest._parse_ch(o.symbol),
				Info: o
			};
		});
	}

	/**
	 * get finished order history
	 */
	async GetTrades(Currency, BaseCurrency, page = 1, pageSize = 200) {
		let data = await this.rest.GetTrades(Currency, BaseCurrency, page, pageSize);
		let trades = [];
		if (data && data.items && data.items.length > 0) {
			data.items.map(o => {
				trades.push({
					Id: o.orderId,
					Time: N.parse(o.createdAt),
					Price: N.parse(o.price),
					Amount: N.parse(o.size),
					DealAmount: N.parse(o.size),
					Status: 'Closed',
					Type: this._order_type(o.side),
					AvgPrice: N.parse(o.price),
					...this.rest._parse_ch(o.symbol),
					Info: o
				});
			});
		}
		return trades;
	}

	async GetPublicTrades(Currency, BaseCurrency) {
		let data = await this.rest.GetPublicTrades(Currency, BaseCurrency);
		let trades = [];
		if (data && data.length > 0) {
			data.map(o => {
				trades.push({
					Id: o.sequence,
					Time: N(o.time).div(1000000).floor(0),
					Price: N.parse(o.price),
					Amount: N.parse(o.size),
					Type: o.side === 'buy' ? 'Buy' : 'Sell',
					Currency: Currency || this.options.Currency,
					BaseCurrency: BaseCurrency || this.options.BaseCurrency,
					Info: o
				});
			});
		};
		return trades;
	}

	Trade(type, price, amount, Currency, BaseCurrency) {
		ok( amount > 0, 'amount should greater than 0');
		if (this.options.Decimals) price = N(price).round(this.options.Decimals);
		if (this.options.StockDecimals) amount = N(amount).floor(this.options.StockDecimals);
		console.log(this.GetName(), type, price, amount, Currency || this.options.Currency, BaseCurrency || this.options.BaseCurrency);
		return this.rest.Trade(type, price, amount, Currency, BaseCurrency);
	}

	Buy(...args) {
		return this.Trade('Buy', ...args);
	}

	Sell(...args) {
		return this.Trade('Sell', ...args);	
	}

	CancelOrder(orderId, Currency, BaseCurrency) {
		return this.rest.CancelOrder(orderId, Currency, BaseCurrency).catch(err => {
			if (err && err.code && err.code * 1 === 400100) return true;
			throw err;
		});
	}

	CancelPendingOrders(Currency, BaseCurrency) {
		console.log('cancelling pending orders...');
		return this.rest.CancelPendingOrders(Currency, BaseCurrency).then(ids => {
			console.log('cancelled orders: ', ids);
			return true;
		});
	}

	GetDepth(Currency, BaseCurrency) {
		return this.rest.GetDepth(Currency, BaseCurrency);
	}

	_order_status( o ) {
		if (o.isActive) return 'Pending';
		return 'Closed';
	}

	_order_type( type ) {
		type = type.toString();
		let arr = {
			'buy': 'Buy',
			'sell': 'Sell'
		};
		return arr[type];
	}

	_funds_type( type ) {
		// 1：币币账户 3：合约账户 6：我的钱包
		type = type.toString();
		let arr = {
			'spot'	:1,
			'future':3,
			'wallet':6
		};
		return arr[type];
	}

	GetMarkets() {
		return this.rest.GetMarkets();
	}

	GetTickers() {
		return this.rest.GetTickers();
	}

}


module.exports = KuCoin;
