const crypto = require('crypto');
const fetch = require('node-fetch');
const N = require('precise-number');
const R = require('ramda');
const debug = require('debug')('exchange:kucoin:rest');
const RateLimit = require('../../lib/rate-limit');
const ExError = require('../../lib/error');
const ErrorCode = require('../../lib/error-code');
const agent = require('../../lib/agent');


class KuCoin_REST {

	constructor(options) {
		this.options = options;
		this.symbol = this._getSymbol();

		this.getAccountRateLimiter = new RateLimit(2000, 6);
		this.tradeRateLimiter = new RateLimit(2000, 20);
		this.cancelRateLimiter = new RateLimit(2000, 20);
		this.orderInfoRateLimiter = new RateLimit(2000, 20);
	}

	_getSymbol(Currency, BaseCurrency) {
		let c = Currency || this.options.Currency;
		let bc = BaseCurrency || this.options.BaseCurrency;
		return `${c}-${bc}`.toUpperCase();
	}

	_parse_ch(ch) {
		if (!ch) ch = '';
		let ms = ch.match(/^([0-9a-z]+)\-([0-9a-z]+)$/i);
		if (ms && ms[1] && ms[2]) {
			return {
				Currency: String(ms[1]).toUpperCase(),
				BaseCurrency: String(ms[2]).toUpperCase()
			};
		} else {
			return {
				Currency: this.options.Currency,
				BaseCurrency: this.options.BaseCurrency
			};
		}
	}

	async fetch(url, params = '', method = 'POST', isPublic = false) {
		await this.options.rateLimiter.wait();

		let headers = {};
		let body = typeof params === 'string' ? params : JSON.stringify(params);
		if (!isPublic) {
			headers['KC-API-KEY'] = this.options.Key;
			let timestamp = Date.now();
			let signBody = `${timestamp}${method}${url}${body}`;
			headers['KC-API-SIGN'] = crypto.createHmac('sha256', this.options.Secret).update(signBody).digest('base64');
			headers['KC-API-TIMESTAMP'] = timestamp;
			headers['KC-API-PASSPHRASE'] = this.options.Pass;
		}

		url = 'https://api.kucoin.com' + url;

		debug('<<', method, url, 'BODY:', body, 'HEADERS:', headers);

		let options = {
			method,
			timeout: method === 'GET' ? 2000 : 5000,
			headers,
			agent: agent.https
		};

		if (method === 'POST' && body) {
			options.body = body;
			options.headers['Content-Length'] = body.length;
			options.headers['Content-Type'] = 'application/json';
		}
		
		return fetch(url, options).then(async res => {
			let status = res.status;
			if (status === 429) throw new ExError(ErrorCode.REQUEST_TOO_FAST, url + ' too many requests');

			let raw = await res.text();
			debug('>>', method, url, 'RESPONSE:', raw);
			if (!raw) throw new Error('okex rest response text is empty');
			let json = null;
			try {
				json = JSON.parse(raw);
			} catch (err) {
				console.error('okex rest method returns bad json:', raw);
			}

			if (json && json.code === '200000') {
				//成功
				return json.data;
			} else {
				let err = new Error(JSON.stringify(json));
				if (!json) json = {};
				err.error_code = json.code || 'unkown';
				err.code = json.code || 'unkown';
				throw err;
			}
		}).catch(err => {
			if (err.type === 'request-timeout') {
				throw new ExError(ErrorCode.REQUEST_TIMEOUT, `rest request (${url}) timeout`, err);
			}
			throw err;
		});
	}

	_getWebsocketInfo() {
		return this.fetch('/api/v1/bullet-public', '', 'POST', true);
	}

	async GetTicker(Currency, BaseCurrency) {
		let data = await this.fetch('/api/v1/market/orderbook/level1?symbol=' + this._getSymbol(Currency, BaseCurrency), null, 'GET', true);
		return {
			High: N.parse(data.bestAsk),
			Low: N.parse(data.bestBid),
			Buy: N.parse(data.bestBid),
			Sell: N.parse(data.bestAsk),
			Last: N.parse(data.price),
			Volume: 0,
			Time: N.parse(data.time),
			Currency: Currency || this.options.Currency,
			BaseCurrency: BaseCurrency || this.options.BaseCurrency
		};
	}

	async GetTickers() {
		let data = await this.fetch('/api/v1/market/allTickers', null, 'GET', true);
		return data.ticker.map(t => {
			return {
				High: N.parse(t.high),
				Low: N.parse(t.low),
				Last: N.parse(t.last),
				Volume: N.parse(t.vol),
				Time: N.parse(data.time),
				...this._parse_ch(t.symbol)
			};
		});
	}

	async GetAccounts(type = '', currency = '') {
		let params = [];
		if (type) params.push('type=' + type);
		if (currency) params.push('currency=' + currency);
		let extra = params.length > 0 ? ('?' + params.join('&')) : '';
		let accounts = await this.fetch('/api/v1/accounts' + extra, '', 'GET');
		/*
		[ { balance: '6996.0721408',
		    available: '6991.809633',
		    holds: '4.26250779',
		    currency: 'USDT',
		    id: '5c718e400d48eb550b244196',
		    type: 'trade' } ]
		 */
		return accounts.map(a => {
			return {
				Currency: a.currency,
				Free: N.parse(a.available),
				Frozen: N.parse(a.holds),
				Info: a
			};
		});
	}

	async GetOrder(orderId) {
		return this.fetch('/api/v1/orders/' + orderId, '', 'GET');
	}

	async GetOrders(Currency, BaseCurrency) {
		let re = await this.fetch('/api/v1/orders?status=active&symbol=' + this._getSymbol(Currency, BaseCurrency), '', 'GET');
		if (re && re.items) {
			return re.items;
		} else {
			throw new Error(JSON.stringify(re));
		}
	}

	//get order history
	GetTrades(Currency, BaseCurrency, page = 1, pageSize = 200) {
		return this.fetch(`/api/v1/fills?currentPage=${page}&pageSize=${pageSize}&symbol=${this._getSymbol(Currency, BaseCurrency)}`, '', 'GET');
	}

	GetPublicTrades(Currency, BaseCurrency) {
		return this.fetch('/api/v1/market/histories?symbol=' + this._getSymbol(Currency, BaseCurrency), null, 'GET');
	}

	async CancelOrder(orderId) {
		return this.fetch('/api/v1/orders/' + orderId, '', 'DELETE').then(r => {
			if (r && r.cancelledOrderIds && r.cancelledOrderIds.indexOf(orderId) !== -1) {
				return true;
			} else {
				throw new Error(JSON.stringify(r));
			}
		});
	}

	async CancelPendingOrders(Currency, BaseCurrency) {
		return this.fetch('/api/v1/orders?symbol=' + this._getSymbol(Currency, BaseCurrency), '', 'DELETE').then(r => {
			if (r && r.cancelledOrderIds) {
				return r.cancelledOrderIds;
			} else {
				throw new Error(JSON.stringify(r));
			}
		});
	}

	async Trade(type, price, amount, Currency, BaseCurrency) {
		let params = {
			clientOid: `${Date.now()}${Math.floor(Math.random() * 1000000)}`,
			symbol: this._getSymbol(Currency, BaseCurrency),
			side: type.toLowerCase(),
			type: 'limit',
			postOnly: this.options.MakerMode ? true : false,
			price,
			size: amount
		};
		let re = await this.fetch('/api/v1/orders', params, 'POST');
		if (re && re.orderId) return re.orderId;
		throw new Error(JSON.stringify(re));
	}

	GetMarkets() {
		return this.fetch('/api/v1/symbols', null, 'GET', true).then(data => {
			let re = [];
			(data || []).map(m => {
				let Currency = m.baseCurrency;
				let BaseCurrency = m.quoteCurrency;
				let Decimals = Math.log10(N(1).div(m.priceIncrement));
				let StockDecimals = Math.log10(N(1).div(m.baseIncrement));
				let MinTradeAmount = N.parse(m.baseMinSize);
				re.push({
					Currency,
					BaseCurrency,
					Decimals,
					StockDecimals,
					MinTradeAmount
				});
			});
			return re;
		});
	}

	GetDepth(Currency, BaseCurrency, size = 20) {
		let symbol = this._getSymbol(Currency, BaseCurrency);
		return this.fetch(`/api/v1/market/orderbook/level2_${size}?symbol=${symbol}`, null, 'GET', true).then(data => {
			if (!data || !data.bids || !data.asks) throw new Error('get kucoin ' + symbol + ' depth error ' + JSON.stringify(data));
			let asks = [];
			let bids = [];

			for (let i = 0;i < data.bids.length;i++) {
				bids.push({
					Price: N.parse(data.bids[i][0]),
					Amount: N.parse(data.bids[i][1])
				});
			}

			for (let i = 0;i < data.asks.length;i++) {
				asks.push({
					Price: N.parse(data.asks[i][0]),
					Amount: N.parse(data.asks[i][1])
				});
			}

			return {
				Asks: R.sort( R.ascend( R.prop('Price') ), asks),
				Bids: R.sort( R.descend( R.prop('Price') ), bids),
				Time: N.parse(data.time),
				Currency: Currency || this.options.Currency,
				BaseCurrency: BaseCurrency || this.options.BaseCurrency,
				Info: {
					sequenceStart: data.sequence,
					sequenceEnd: data.sequence
				}
			};
		});
	}


	_getOrderBook(Currency, BaseCurrency) {
		let symbol = this._getSymbol(Currency, BaseCurrency);
		return this.fetch(`/api/v1/market/orderbook/level2_100?symbol=${symbol}`, null, 'GET', true).then(data => {
			if (!data || !data.bids || !data.asks) throw new Error('get kucoin ' + symbol + ' depth error ' + JSON.stringify(data));
			let asks = {};
			let bids = {};

			for (let i = 0;i < data.bids.length;i++) {
				bids[data.bids[i][0]] = N.parse(data.bids[i][1]);
			}

			for (let i = 0;i < data.asks.length;i++) {
				asks[data.asks[i][0]] = N.parse(data.asks[i][1]);
			}

			return {
				Asks: asks,
				Bids: bids,
				Time: N.parse(data.time),
				Currency: Currency || this.options.Currency,
				BaseCurrency: BaseCurrency || this.options.BaseCurrency,
				sequenceStart: data.sequence,
				sequenceEnd: data.sequence,
				StartTime: Date.now()
			};
		});
	}
}

module.exports = KuCoin_REST;
