const N = require('precise-number');
const WebSocket = require('../../lib/auto-reconnect-ws.js');
const moment = require('moment');
const R = require('ramda');
const Events = require('events');
const pako = require('pako');

class KuCoin_WS extends Events {
	constructor(options, rest) {
		super();
		this.options = options;
		this.rest = rest;

		this.lastPingTime = Date.now();

		this.symbol = this.options.Currency.toUpperCase() + '-' + this.options.BaseCurrency.toUpperCase();
		this.orderbooks = {};
		this.orderbooksCache = {};
		this.orderbooksGetting = {};

		this.boot().catch(err => {
			console.error('kucoin websocket boot error:', err);
		});
	}

	async boot() {
		let info = await this.rest._getWebsocketInfo();
		if (!info || !info.instanceServers || !info.instanceServers[0]) throw new Error(JSON.stringify(info));

		let url = `${info.instanceServers[0].endpoint}?token=${encodeURIComponent(info.token)}&[connectId=fb${Math.round(Math.random() * 1000000)}]`;
		this.ws = new WebSocket(url);

		//remember subscription commands
		this.subscriptionCommands = [];

		if (this.options.onDepth) {
			this.subscriptionCommands.push({
				type: "subscribe",
				topic: "/market/level2:" + this.symbol,
				response: true
			});
		}

		if (this.options.onTicker) {
			this.subscriptionCommands.push({
				type: "subscribe",
				topic: "/market/ticker:" + this.symbol,
				response: true
			});
		}

		if (this.options.onPublicTrades) {
			this.subscriptionCommands.push({
				type: 'subscribe',
				topic: `/market/match:${this.symbol}`,
				privateChannel: false,
				response: true
			});
		}

		this.wsReady = false;
		this.ws.on('open', () => {
			this.emit('connect');
			console.log('kucoin websocket connected');
			this.subscriptionCommands.map(cmd => {
				cmd.id = `${Date.now()}${Math.round(Math.random() * 1000000)}`;
				this.ws.send(JSON.stringify(cmd));
			});
		});

		this.ws.on('message', (data) => {
			this.wsReady = true;
			this.onMessage(data);
		});

		this.ws.on('close', () => {
			this.emit('close');
			console.log('kucoin websocket closed');
			this.wsReady = false;
		});

		this.ws.on('error', () => {
			this.emit('close');
			console.log('kucoin websocket error');
			this.wsReady = false;
		});

		setInterval(() => {
			if (this.wsReady) {
				this.lastPingTime = Date.now();
				this.ws.send(JSON.stringify({
					id: `${Date.now()}`,
					type: 'ping'
				}));
			}
		}, 11000);
	}

	onMessage(raw)  {
		if (typeof raw !== 'string') {
			raw = pako.inflateRaw(raw, {to: 'string'});
		}

		let msg = null;
		try {
			msg = JSON.parse(raw);
		} catch (err) {
			console.error('okex websocket message error: ' + raw);
			return;
		}
		if (!msg) return;

		if (msg && msg.type === 'pong') {
			this.onPong(msg);
			return;
		}

		if (msg && msg.subject === 'trade.l2update') {
			this.onOrderBookUpdate(msg);
		} else if (msg && msg.subject === 'trade.ticker') {
			this.onTicker(msg);
		} else if (msg && msg.subject === 'trade.l3match') {
			this.onPublicTrades(msg);
		}
	}

	onPong() {
		let gap = Date.now() - this.lastPingTime;
		console.log('kucoin websocket ping pong time:', gap + 'ms');
		this.lastPong = Date.now();
	}

	addSubscription(Currency, BaseCurrency, type) {
		let json = null;
		let symbol = this._getSymbol(Currency, BaseCurrency);

		if (type === 'Ticker') {
			json = {
				type: "subscribe",
				topic: "/market/ticker:" + symbol,
				response: true
			};
		} else if (type === 'Depth') {
			json = {
				type: "subscribe",
				topic: "/market/level2:" + symbol,
				response: true
			};
		} else if (type === 'PublicTrades') {
			json = {
				type: 'subscribe',
				topic: `/market/match:${symbol}`,
				privateChannel: false,
				response: true
			};
		}

		if (!json) throw new Error('unkown subscription type: ' + type);

		this.ws.send(JSON.stringify(json));
		this.subscriptionCommands.push(json);
	}

	async rebuildOrderBook(symbol) {
		let { Currency, BaseCurrency } = this._parse_ch(symbol);
		if (this.orderbooksGetting[symbol]) return;
		console.log('rebuilding order book for', symbol);
		this.orderbooksGetting[symbol] = true;
		try {
			this.orderbooks[symbol] = await this.rest._getOrderBook(Currency, BaseCurrency);
			this.playBackOrderBookUpdates(symbol);
		} catch (err) {
			console.error('rebuild order book error:', err);
		}
		this.orderbooksGetting[symbol] = false;
	}

	playBackOrderBookUpdates(symbol) {
		let updates = this.orderbooksCache[symbol];
		delete this.orderbooksCache[symbol];
		if (updates && updates.length > 0) {
			console.log('play back order book updates for', symbol, updates.length);
			updates.map(update => {
				this.updateOrderBook(update);
			});
		}
	}

	onOrderBookUpdate(msg) {
		let { data: { symbol, sequenceStart, sequenceEnd, changes } } = msg;
		if (!symbol || !changes || !sequenceStart || !sequenceEnd) return;
		if (!this.orderbooks[symbol]) {
			if (!this.orderbooksCache[symbol]) this.orderbooksCache[symbol] = [];
			this.orderbooksCache[symbol].push(msg);
			this.rebuildOrderBook(symbol);
		} else {
			this.updateOrderBook(msg);
		}
	}

	updateOrderBook(msg) {
		let { data: { symbol, sequenceStart, sequenceEnd, changes } } = msg;
		let orderbook = this.orderbooks[symbol];
		if (!orderbook) return;

		if (orderbook && orderbook.sequenceStart && orderbook.sequenceEnd) {
			if (sequenceStart > orderbook.sequenceEnd) {
				let hasChange = false;
				let gap = sequenceStart - orderbook.sequenceEnd;
				// console.log('l2update gap', gap);

				if (gap !== 1 || Date.now() - orderbook.StartTime > 60000) {
					delete this.orderbooks[symbol];
					this.onOrderBookUpdate(msg);
					return;
				}

				if (changes.asks && changes.asks.length > 0) {
					changes.asks.map(ask => {
						let [price, size] = ask;
						size = N.parse(size);
						if (N.parse(price) > 0) {
							if (size === 0) {
								delete orderbook.Asks[price];
							} else {
								orderbook.Asks[price] = size;
							}
							hasChange = true;
						}
					});
				}
				if (changes.bids && changes.bids.length > 0) {
					changes.bids.map(bid => {
						let [price, size] = bid;
						size = N.parse(size);
						if (N.parse(price) > 0) {
							if (size === 0) {
								delete orderbook.Bids[price];
							} else {
								orderbook.Bids[price] = size;
							}
							hasChange = true;
						}
					});
				}
				orderbook.sequenceEnd = sequenceEnd;

				if (hasChange) {
					let asks = Object.keys(orderbook.Asks).map(price => {
						return {
							Price: N.parse(price),
							Amount: orderbook.Asks[price]
						};
					});
					let bids = Object.keys(orderbook.Bids).map(price => {
						return {
							Price: N.parse(price),
							Amount: orderbook.Bids[price]
						};
					});

					let depth = {
						Asks: R.sort( R.ascend( R.prop('Price') ), asks).slice(0, 30),
						Bids: R.sort( R.descend( R.prop('Price') ), bids).slice(0, 30),
						Time: Date.now(),
						...this._parse_ch(symbol)
					};

					if (depth.Asks.length === 0 || depth.Bids.length === 0 || depth.Asks[0].Price <= depth.Bids[0].Price) {
						delete this.orderbooks[symbol];
						this.onOrderBookUpdate(msg);
						return;
					}

					if (this.options.onDepth) this.options.onDepth(depth);
				}
			} else {
				// console.log('ignore old update', msg);
			}
		}
	}

	_getSymbol(Currency, BaseCurrency) {
		let c = Currency || this.options.Currency;
		let bc = BaseCurrency || this.options.BaseCurrency;
		return `${c}-${bc}`.toUpperCase();
	}

	_parse_ch(ch) {
		if (!ch) ch = '';
		let ms = ch.match(/^([0-9a-z]+)\-([0-9a-z]+)$/i);
		if (ms && ms[1] && ms[2]) {
			return {
				Currency: String(ms[1]).toUpperCase(),
				BaseCurrency: String(ms[2]).toUpperCase()
			};
		} else {
			return {
				Currency: this.options.Currency,
				BaseCurrency: this.options.BaseCurrency
			};
		}
	}

	onTicker(msg) {
		if (!msg || !msg.data || !msg.topic) return;
		let symbol = msg.topic.replace('/market/ticker:', '');
		let data = msg.data;
		let re = {
			Buy: N.parse(data.bestBid),
			Sell: N.parse(data.bestAsk),
			Volume: 0,
			Last: N.parse(data.price),
			Time: N.parse(data.time),
			...this._parse_ch(symbol),
			Info: data
		};
		if (this.options.onTicker) this.options.onTicker(re);
	}

	onPublicTrades(msg) {
		let { data: {symbol, side, size, price, time, tradeId} } = msg;
		if (!symbol || !side || !size || !price || !time || !tradeId) return;

		let trade = {
			Amount: N.parse(size),
			Price: N.parse(price),
			Type: side === 'buy' ? 'Buy' : 'Sell',
			Time: N.parse(`${time}`.substr(0, 13)),
			Id: tradeId,
			...this._parse_ch(symbol)
		};
		if (this.options.onPublicTrades) this.options.onPublicTrades([trade]);
	}
}

module.exports = KuCoin_WS;
