const EXCHANGE_REST = require('./rest.js');
const EXCHANGE_WS = require('./ws.js');
const N = require('precise-number');
const { ok } = require('assert');
const EXCHANGE = require('../exchange.js');
const ExError = require('../../lib/error');
const ErrorCode = require('../../lib/error-code');
const fetch = require('node-fetch');

class OKEX extends EXCHANGE {
	constructor(options) {
		options = Object.assign({
			Name: 'OKEX',
			Fees: {
				Maker: 0.002,
				Taker: 0.002
			},
			RateLimit: 10
		}, options);
		super(options);

		this.rest = new EXCHANGE_REST(this.options);
	
		if (options.isWS) {
			this.ws = new EXCHANGE_WS(options);
			this.ws.on('connect', () => {
				this.wsReady = true;
			});
			this.ws.on('close', () => {
				this.wsReady = false;
			});
		}
	}

	async Subscribe(currency, baseCurrency, type) {
		if (!this.options.isWS) throw new Error('is not websocket mode');
		if (['Depth', 'PublicTrades', 'Ticker'].indexOf(type) === -1) {
			throw new Error('unkown subscription type: ' + type);
		}

		if (type === 'Depth' && !this.options.onDepth) throw new Error('no onDepth callback');
		if (type === 'Ticker' && !this.options.onTicker) throw new Error('no onTicker callback');
		if (type === 'PublicTrades' && !this.options.onPublicTrades) throw new Error('no onPublicTrades callback');

		try {
			await this.waitUntilWSReady();
			await this.ws.addSubscription(currency, baseCurrency, type);
		} catch (err) {
			console.error(this.GetName() + ` Subscribe got error:`, err);
			throw err;
		}
	}

	GetAccount(...args) {
		return this.rest.GetAccount(...args);
	}

	GetAccounts(...args) {
		return this.rest.GetAccounts(...args);
	}

	GetTicker(Currency, BaseCurrency) {
		return this.rest.GetTicker(Currency, BaseCurrency);
	}

	GetOrder(...args) {
		return this.rest.GetOrder(...args);
	}

	GetTrades(Currency, BaseCurrency) {
		return this.rest.GetTrades(Currency, BaseCurrency);
	}

	GetPublicTrades(Currency, BaseCurrency, since) {
		return this.rest.GetPublicTrades(Currency, BaseCurrency, since);
	}

	Trade(type, price, amount, Currency, BaseCurrency, ClientId) {
		ok( amount > 0, 'amount should greater than 0');
		if (this.options.Decimals) price = N(price).round(this.options.Decimals);
		if (this.options.StockDecimals) amount = N(amount).floor(this.options.StockDecimals);
		console.log(this.GetName(), type, price, amount, Currency || this.options.Currency, BaseCurrency || this.options.BaseCurrency);
		return this.rest.Trade(type, price, amount, Currency, BaseCurrency, ClientId).then( r => {
			if (r && r.result && r.order_id) {
				return r.order_id;
			} else {
				throw new Error(JSON.stringify(r));
			}
		}).catch(err => {
			if (err && err.error_code === 33017) {
				throw new ExError(ErrorCode.INSUFFICIENT_BALANCE, err.message, err);
			} else {
				throw err;
			}
		});
	}

	Buy(...args) {
		return this.Trade('Buy', ...args);
	}

	Sell(...args) {
		return this.Trade('Sell', ...args);	
	}

	CancelOrder(orderId, Currency, BaseCurrency) {
		return this.rest.CancelOrder(orderId, Currency, BaseCurrency).then(result => {
			if (result && (result.result === true || result.result === 'true')) return true;
			throw new Error('cancel order failed, result: ' + JSON.stringify(result));
		}).catch(err => {
			if (err && err.error_code === 33027) return true;
			if (err && err.error_code === 33014) return true;
			throw err;
		});
	}

	CancelPendingOrders(Currency, BaseCurrency) {
		console.log('cancelling pending orders...');
		return this.GetOrders(Currency, BaseCurrency).then( orders => {
			console.log('cancelling', orders.length, 'orders');
			return Promise.all(orders.map( o => {
				return this.CancelOrder(o.Id, Currency, BaseCurrency);
			})).then( results => {
				console.log(results);
				return true;
			});
		});
	}

	FundsTransfer(Currency, Amount, From, To) {
		From = this._funds_type(From);
		To 	 = this._funds_type(To);
		if (!From || !To) {
			throw new Error('Wrong funds type,[spot,future,wallet] expected');
		}
		return this.rest.FundsTransfer(Currency, Amount, From, To);
	}

	GetOrders(Currency, BaseCurrency) {
		return this.rest.GetOrders(Currency, BaseCurrency);
	}

	GetDepth(...args) {
		return this.rest.GetDepth(...args);
	}



	_funds_type( type ) {
		// 1：币币账户 3：合约账户 6：我的钱包
		type = type.toString();
		let arr = {
			'spot'	:1,
			'future':3,
			'wallet':6
		};
		return arr[type];
	}

	GetMarkets() {
		return this.rest.GetMarkets();
	}

	GetTickers() {
		return this.rest.GetTickers();
	}

	GetRecords(minutes, Currency, BaseCurrency) {
		return this.rest.GetRecords(minutes, Currency, BaseCurrency);
	}

}


module.exports = OKEX;
