const fetch = require('node-fetch');
const N = require('precise-number');
const R = require('ramda');
const debug = require('debug')('exchange:okex:rest');
const agent = require('../../lib/agent');
const crypto = require('crypto');

class OKEX {

	constructor(options) {
		this.options = options;
		this.symbol = this._getSymbol(options.Currency, options.BaseCurrency);
		if (!this.options.rateLimiter) throw 'No rateLimiter in options';
	}

	async fetch(method, path, postBody = '', isPublic = false) {
		await this.options.rateLimiter.wait();
		if (method === 'POST') {
			if (typeof postBody === 'object') postBody = JSON.stringify(postBody);
		}

		debug('<<', method, 'https://www.okex.com' + path, 'BODY:', postBody);

		let timestamp = new Date().toISOString();
		let headers = {
			'Content-Type': 'application/json'
		};
		if (!isPublic) {
			headers['OK-ACCESS-KEY'] = this.options.Key;
			headers['OK-ACCESS-SIGN'] = this.sign(timestamp, method, path, postBody);
			headers['OK-ACCESS-TIMESTAMP'] = timestamp;
			headers['OK-ACCESS-PASSPHRASE'] = this.options.Pass;
		}

		let options = {
			method,
			timeout: method === 'GET' ? 4000 : 4000,
			headers,
			agent: agent.https
		};

		if (postBody) options.body = postBody;

		return fetch('https://www.okex.com' + path, options).then(res => res.text()).then(t => {
			debug('>> ' + t);

			if (!t) return Promise.reject('OKEX returns empty: ' + path);
			try {
				let d = JSON.parse(t);
				return Promise.resolve(d);
			} catch ( err ) {
				return Promise.reject('OKEX JSON parse error: ' + t);
			}
		}).then(json => {

			if (json && json.error_message && json.error_code) {
				let err = new Error(json.error_message);
				err.error_code = json.error_code * 1;
				err.original = json;
				err.options = R.pick(['headers', 'body', 'method'])(options),
				err.path = path;
				throw err;
			}

			if (json && json.code && json.message) {
				let err = new Error(json.message);
				err.error_code = json.code * 1;
				err.original = json;
				err.options = R.pick(['headers', 'body', 'method'])(options),
				err.path = path;
				throw err;
			}

			if (json && json.message === 'System error') {
				throw new Error('OKEX: System error');
			} else {
				return Promise.resolve(json);
			}
		}).catch(err => {
			if (err && err.type === 'request-timeout') {
				throw path + ' timeout';
			}
			throw err;
		});
	}

	sign(timestamp, method, path, body) {
		let presign = timestamp + method + path;
		if (body) presign += body;
		return crypto.createHmac('sha256', this.options.Secret).update(presign).digest('base64');
	}

	_getSymbol(Currency, BaseCurrency) {
		if (!Currency) Currency = this.options.Currency;
		if (!BaseCurrency) BaseCurrency = this.options.BaseCurrency;
		return `${Currency}-${BaseCurrency}`;
	}

	_parseSymbol(Currency, BaseCurrency) {
		if (!Currency) Currency = this.options.Currency;
		if (!BaseCurrency) BaseCurrency = this.options.BaseCurrency;
		return { Currency, BaseCurrency };
	}

	_parseSymbolString(symbol) {
		if (!symbol) symbol = this.symbol;
		let [Currency, BaseCurrency] = String(symbol).toUpperCase().split('-');
		if (!Currency) Currency = this.options.Currency;
		if (!BaseCurrency) BaseCurrency = this.options.BaseCurrency;
		return { Currency, BaseCurrency };
	}

	GetTicker(Currency, BaseCurrency) {
		return this.fetch('GET', `/api/spot/v3/instruments/${this._getSymbol(Currency, BaseCurrency)}/ticker`, null, true).then(data => this._parseTicker(data));
	}

	_parseTicker(data) {
		return {
			High: N.parse(data.high_24h),
			Low: N.parse(data.low_24h),
			Buy: N.parse(data.best_bid),
			Sell: N.parse(data.best_ask),
			Last: N.parse(data.last),
			Volume: N.parse(data.volume_24h),
			Time: new Date(data.timestamp).getTime(),
			...this._parseSymbolString(data.instrument_id)
		};
	}

	GetTickers() {
		return this.fetch('GET', `/api/spot/v3/instruments/ticker`, null, true).then(data => data.map(t => this._parseTicker(t)));
	}

	GetTrades(Currency, BaseCurrency) {
		return this.fetch('GET', `/api/spot/v3/orders?instrument_id=${this._getSymbol(Currency, BaseCurrency)}&state=7`).then(arr => arr.map(t => this._parseOrder(t))).then(arr => arr.filter(o => o.DealAmount > 0));
	}

	async GetPublicTrades(Currency, BaseCurrency) {
		let data = await this.fetch('GET', `/api/spot/v3/instruments/${this._getSymbol(Currency, BaseCurrency)}/trades`, null, true);
		let trades = [];
		if (data && data.length > 0) {
			data.map(o => {
				trades.push({
					Id: o.trade_id,
					Time: new Date(o.timestamp).getTime(),
					Price: N.parse(o.price),
					Amount: N.parse(o.size),
					Type: o.side === 'buy' ? 'Buy' : 'Sell',
					...this._parseSymbol(Currency, BaseCurrency)
				});
			});
		};
		return trades;
	}

	async GetAccount(Currency, BaseCurrency) {
		if (!Currency) Currency = this.options.Currency;
		if (!BaseCurrency) BaseCurrency = this.options.BaseCurrency;
		let map = (await this.GetAccounts()).reduce((map, a) => {
			map[a.Currency] = a;
			return map;
		}, {});
		return {
			Balance: map[BaseCurrency] ? map[BaseCurrency].Free : 0,
			FrozenBalance: map[BaseCurrency] ? map[BaseCurrency].Frozen : 0,
			Stocks: map[Currency] ? map[Currency].Free : 0,
			FrozenStocks: map[Currency] ? map[Currency].Frozen : 0
		};
	}

	GetAccounts(type = 'spot') {
		let path = `/api/spot/v3/accounts`;
		if (type === 'wallet') path = '/api/account/v3/wallet'; 
		return this.fetch('GET', path, null).then(data => {
			return data.map(a => {
				return {
					Currency: String(a.currency).toUpperCase(),
					Free: N.parse(a.available),
					Frozen: N.parse(a.hold),
					Info: a
				};
			});
		});
	}

	GetOrder(orderId, Currency, BaseCurrency) {
		return this.fetch('GET', `/api/spot/v3/orders/${orderId}?instrument_id=${this._getSymbol(Currency, BaseCurrency)}`).then(o => {
			return this._parseOrder(o);
		});
	}

	async GetOrders(Currency, BaseCurrency) {
		 let orders = await this.fetch('GET', `/api/spot/v3/orders_pending?instrument_id=${this._getSymbol(Currency, BaseCurrency)}`);
		 return orders.map(o => this._parseOrder(o));
	}

	CancelOrder(order_id, Currency, BaseCurrency) {
		return this.fetch('POST', `/api/spot/v3/cancel_orders/${order_id}`, {
			instrument_id: this._getSymbol(Currency, BaseCurrency),
			order_id
		});
	}

	Trade(direction, price, amount, Currency, BaseCurrency, ClientId = '') {
		let params = {
			instrument_id: this._getSymbol(Currency, BaseCurrency),
			side: direction.toLowerCase(),
			order_type: this.options.MakerMode ? 1 : 0,
			price,
			size: amount,
			match_price: price === -1 ? '1' : '0',
			lever_rate: this.options.MarginLevel
		};
		if (ClientId) params.client_oid = ClientId;
		return this.fetch('POST', `/api/spot/v3/orders`, params);
	}

	GetDepth(Currency, BaseCurrency, size) {
		if (!size) size = 30;
		return this.fetch('GET', `/api/spot/v3/instruments/${this._getSymbol(Currency, BaseCurrency)}/book?size=${size}`, null, true).then(data => {

			if (!data || !data.bids || !data.asks) throw new Error('get okex ' + this.symbol + ' depth error ' + JSON.stringify(data));

			let asks = [];
			let bids = [];

			for (let i = 0;i < data.bids.length;i++) {
				bids.push({
					Price: data.bids[i][0] * 1,
					Amount: data.bids[i][1] * 1
				});
			}

			for (let i = 0;i < data.asks.length;i++) {
				asks.push({
					Price: data.asks[i][0] * 1,
					Amount: data.asks[i][1] * 1
				});
			}

			bids.sort((a, b) => {
				let diff = a.Price - b.Price;
				return diff > 0 ? -1 : diff === 0 ? 0 : 1;
			});

			asks.sort((a, b) => {
				let diff = a.Price - b.Price;
				return diff > 0 ? 1 : diff === 0 ? 0 : -1;
			});

			return Promise.resolve({
				Asks: R.sort( R.ascend( R.prop('Price') ), asks),
				Bids: R.sort( R.descend( R.prop('Price') ), bids),
				Time: new Date(data.timestamp).getTime(),
				...this._parseSymbol(Currency, BaseCurrency)
			});
		});
	}

	_parseOrder(o) {
		return {
			Id: o.order_id,
			ClientId: o.client_oid,
			Time: new Date(o.timestamp).getTime(),
			Type: o.side === 'sell' ? 'Sell' : 'Buy',
			Amount: N.parse(o.size),
			Price: N.parse(o.price),
			AvgPrice: N.parse(o.price),
			DealAmount: N.parse(o.filled_size),
			...this._parseSymbolString(o.instrument_id),
			Status: this._order_status(o.state),
			Info: o
		};
	}

	_order_status( status ) {
		//-1:已撤销  0:未成交  1:部分成交  2:完全成交 3:撤单处理中
		status = String(status);
		let arr = {
			'-2': 'Closed',
			'-1':'Cancelled',
			'0': 'Pending',
			'1': 'Partial',
			'2': 'Closed',
			'3': 'Pending',
			'4': 'Cancelled'
		};
		return arr[status];
	}

	async GetMarkets() {
		let arr = await this.fetch('GET', `/api/spot/v3/instruments`, null, true);
		return arr.map(a => {
			return {
				Currency: a.base_currency,
				BaseCurrency: a.quote_currency,
				Decimals: Math.log10(N(1).div(a.tick_size)),
				StockDecimals: Math.log10(N(1).div(a.size_increment)),
				MinTradeAmount: N.parse(a.min_size)
			};
		});
	}

	async GetRecords(minutes = 1, Currency, BaseCurrency) {
		let arr = await this.fetch('GET', `/api/spot/v3/instruments/${this._getSymbol(Currency, BaseCurrency)}/candles?granularity=${minutes * 60}`, null, true);
		return arr.map(k => {
			let [timestamp, open, high, low, close, volume] = k;
			return {
				Open: N.parse(open),
				High: N.parse(high),
				Low: N.parse(low),
				Close: N.parse(close),
				Volume: N.parse(volume),
				Time: new Date(timestamp).getTime(),
				CurrencyVolume: volume * close
			};
		}).sort((a, b) => {
			return a.Time - b.Time;
		}).slice(-200);
	}
}

module.exports = OKEX;
