const N = require('precise-number');
const WebSocket = require('../../lib/auto-reconnect-ws.js');
const crypto = require('crypto');
const R = require('ramda');
const Events = require('events');
const tc = require('turbocolor');
const pako = require('pako');

class OKEXV3_WS extends Events {
	constructor(options) {
		super();
		this.options = options;

		this.lastPingTime = Date.now();

		this.symbol = `${this.options.Currency}-${this.options.BaseCurrency}`;

		this.ws = new WebSocket('wss://real.okex.com:8443/ws/v3');

		//remember subscription commands
		this.subscriptionCommands = [];

		if (this.options.onDepth) {
			this.subscriptionCommands.push(JSON.stringify({
				op: 'subscribe',
				args: [`spot/depth5:${this.symbol}`]
			}));
		}

		if (this.options.onTicker) {
			this.subscriptionCommands.push(JSON.stringify({
				op: 'subscribe',
				args: [`spot/ticker:${this.symbol}`]
			}));
		}

		if (this.options.onPublicTrades) {
			this.subscriptionCommands.push(JSON.stringify({
				op: 'subscribe',
				args: [`spot/trade:${this.symbol}`]
			}));
		}

		this.wsReady = false;
		this.ws.on('open', () => {
			this.subscriptionCommands.map(cmd => {
				this.ws.send(cmd);
			});
			this.emit('connect');
		});

		this.ws.on('message', (data) => {
			this.wsReady = true;
			this.onMessage(data);
		});

		this.ws.on('pong', t => {
			this.wsReady = true;
			if (this.options.onPong) {
				this.options.onPong(t);
			}
		});

		this.ws.on('close', () => {
			this.emit('close');
			this.wsReady = false;
		});

		this.ws.on('error', () => {
			this.emit('close');
			this.wsReady = false;
		});

		setInterval(() => {
			if (this.wsReady) {
				this.lastPingTime = Date.now();
				this.ws.send('ping');
			}
		}, 5000);
	}

	sign(timestamp, method, path, body) {
		let presign = timestamp + method + path;
		if (body) presign += body;
		return crypto.createHmac('sha256', this.options.Secret).update(presign).digest('base64');
	}

	onMessage(raw)  {
		if (typeof raw !== 'string') {
			raw = pako.inflateRaw(raw, {to: 'string'});
		}
		if (raw === 'pong') {
			this.onPong();
			return;
		}

		let message = null;
		try {
			message = JSON.parse(raw);
		} catch (err) {
			console.error('okex websocket message error: ' + raw);
			return;
		}
		if (!message) return;

		if (message.event === 'login') {
			if (message.success) {
				console.log(tc.green('okex websocket login success'));
				this.wsReady = true;
				this.emit('connect');
			} else {
				console.error(tc.red('okex websocket login failed'), message);
				process.exit();
				return;
			}
			return;
		}

		if (message.event === 'error') {
			console.error(message);
			return;
		}

		let { table, data } = message;
		if (!table || !data) return;

		if (table === 'spot/ticker') {
			this.onTicker(message);
		} else if (table === 'spot/depth5') {
			this.onDepth(message);
		} else if (table === 'spot/trade') {
			this.onPublicTrades(message);
		}
	}

	onPong() {
		// let gap = Date.now() - this.lastPingTime;
		// console.log('okex websocket ping pong time:', gap + 'ms');
		this.lastPong = Date.now();
	}

	addSubscription(Currency, BaseCurrency, type) {
		if (!Currency) Currency = this.options.Currency;
		if (!BaseCurrency) BaseCurrency = this.options.BaseCurrency;

		let channel = '';
		if (type === 'Ticker') {
			channel = `spot/ticker:${this._getSymbol(Currency, BaseCurrency)}`;
		} else if (type === 'Depth') {
			channel = `spot/depth5:${this._getSymbol(Currency, BaseCurrency)}`;
		} else if (type === 'PublicTrades') {
			channel = `spot/trade:${this._getSymbol(Currency, BaseCurrency)}`;
		}

		if (!channel) throw new Error('unkown subscription type: ' + type);

		let cmd = JSON.stringify({
			op: 'subscribe',
			args: [channel]
		});

		this.ws.send(cmd);
		this.subscriptionCommands.push(cmd);
	}

	onDepth(data) {
		data = data.data[0];
		let {timestamp, asks, bids, instrument_id} = data;
		if (!asks || !bids || !instrument_id || !timestamp) return;

		asks = asks.map(pair => {
			return {
				Price: N.parse(pair[0]),
				Amount: N.parse(pair[1])
			};
		});

		bids = bids.map(pair => {
			return {
				Price: N.parse(pair[0]),
				Amount: N.parse(pair[1])
			};
		});

		let depth = {
			Time: new Date(timestamp).getTime(),
			Asks: R.sort( R.ascend( R.prop('Price') ), asks),
			Bids: R.sort( R.descend( R.prop('Price') ), bids),
			...this._parse_ch(instrument_id)
		};

		this.options.onDepth(depth);
	}

	_parse_ch(symbol) {
		if (!symbol) symbol = this.symbol;
		let [Currency, BaseCurrency] = String(symbol).toUpperCase().split('-');
		if (!Currency) Currency = this.options.Currency;
		if (!BaseCurrency) BaseCurrency = this.options.BaseCurrency;
		return { Currency, BaseCurrency };
	}

	_getSymbol(Currency, BaseCurrency) {
		if (!Currency) Currency = this.options.Currency;
		if (!BaseCurrency) BaseCurrency = this.options.BaseCurrency;
		return `${Currency}-${BaseCurrency}`.toUpperCase();
	}

	onTicker({data}) {
		if (!data || data.length <= 0) return;
		let tick = data[0];
		let re = {
			High: N.parse(tick.high_24h),
			Low: N.parse(tick.low_24h),
			Close: N.parse(tick.last),
			Last: N.parse(tick.last),
			Volume: N.parse(tick.volume_24h),
			Time: new Date(tick.timestamp).getTime(),
			...this._parse_ch(tick.instrument_id)
		};
		if (this.options.onTicker) {
			this.options.onTicker(re);
		}
	}

	onPublicTrades({data}) {
		if (!data || data.length <= 0) return;
		let trades = data.map(t => {
			return {
				Amount: N.parse(t.size),
				Price: N.parse(t.price),
				Type: t.side === 'buy' ? 'Buy' : 'Sell',
				Time: new Date(t.timestamp).getTime(),
				Id: t.trade_id,
				...this._parse_ch(t.instrument_id)
			};
		});
		this.options.onPublicTrades(trades);
	}
}

module.exports = OKEXV3_WS;
