const EXCHANGE_REST = require('./rest.js');
const EXCHANGE_WS = require('./ws.js');
const N = require('precise-number');
const { ok } = require('assert');
const EXCHANGE = require('../exchange.js');

class OKEXV3_FUTURE extends EXCHANGE {
	constructor(options) {
		options = Object.assign({
			Name: 'OKEXV3_FUTURE',
			Fees: {
				Maker: 0.0001,
				Taker: 0.0003
			},
			RateLimit: 10,
			Decimals: 8,
			StockDecimals: 3,
			BaseCurrency: 'USD',
			Currency: 'BTC',
			ContractType: 'quarter',
			ContractMode: 'All',
			MarginLevel: 10,
			MinTradeAmount: 1
		}, options);
		super(options);

		ok(options.Pass, 'no pass phrase');

		if (['All', 'Seperate'].indexOf(options.ContractMode) === -1) throw new Error('options.ContractMode should either be All or Seperate');
		if (options.MarginLevel !== 10 && options.MarginLevel !== 20) throw new Error('wrong MarginLevel');

		this.rest = new EXCHANGE_REST(this.options);

		if (options.isWS) {
			this.ws = new EXCHANGE_WS(this.options, this.rest);
			this.ws.on('connect', () => {
				this.wsReady = true;
			});
			this.ws.on('close', () => {
				this.wsReady = false;
			});
		}
	}

	_check_contract_type(contract_type) {
		if (['this_week', 'next_week', 'quarter'].indexOf(contract_type) === -1) {
			throw new Error('unkown contract_type: ' + contract_type);
		}
	}

	_on_trade(cb) {
		return (err, o) => {
			if (err) {
				cb(err, o);
				return;
			}
			cb(false, {
				Id: o.orderId,
				Price: N.parse(o.tradePrice),
				AvgPrice: N.parse(o.averagePrice),
				Amount: N.parse(o.tradeAmount),
				DealAmount: N.parse(o.completedTradeAmount),
				Status: this._order_status(o.status),
				Type: this._order_type(o.tradeType),
				LeftAmount: N.parse(o.unTrade),
				UnitPrice: N.parse(o.tradeUnitPrice),
				SigAmount: N.parse(o.sigTradeAmount),
				SigPrice: N.parse(o.sigTradePrice),
				Time: N.parse(o.createdDate)
			});
		};
	}

	_on_account_change(cb) {
		return (err, data) => {
			if (err) {
				cb(err, data);
				return;
			}
			let currency = this.options.Currency.toLowerCase();
			cb(false, {
				Balance: N.parse(data.info.free.usd),
				FrozenBalance: N.parse(data.info.freezed.usd),
				Stocks: N.parse(data.info.free[currency]),
				FrozenStocks: N.parse(data.info.freezed[currency])
			});
		};
	}

	async Subscribe(Currency, BaseCurrency, ContractType, type) {
		if (!this.options.isWS) throw new Error('is not websocket mode');
		if (['Depth', 'PublicTrades', 'Ticker'].indexOf(type) === -1) {
			throw new Error('unkown subscription type: ' + type);
		}

		this._check_contract_type(ContractType);

		if (type === 'Depth' && !this.options.onDepth) throw new Error('no onDepth callback');
		if (type === 'Ticker' && !this.options.onTicker) throw new Error('no onTicker callback');
		if (type === 'PublicTrades' && !this.options.onPublicTrades) throw new Error('no onPublicTrades callback');

		try {
			await this.waitUntilWSReady();
			await this.ws.addSubscription(Currency, BaseCurrency, ContractType, type);
		} catch (err) {
			console.error(this.GetName() + ` Subscribe got error:`, err);
			throw err;
		}
	}

	GetPublicTrades(...args) {
		return this.rest.GetPublicTrades(...args);
	}

	GetTrades() {
		throw new Error('not support');
	}

	API(url, params, method) {
		if (!method) method = 'POST';
		return this.rest.fetch(url, params, method);
	}

	//设置杠杆倍率
	SetMarginLevel(level) {
		this.options.MarginLevel = level;
	}

	getHandler() {
		return this.rest;
	}

	GetAccount(...args) {
		return this.rest.GetAccount(...args);
	}

	async GetAccounts() {
		let markets = await this.GetMarkets();
		let coins = markets.reduce((acc, m) => {
			acc[m.Currency] = m.Currency;
			return acc;
		}, {});
		let re = [];
		for (let coin of Object.keys(coins)) {
			let a = await this.GetAccount(coin);
			re.push({
				Currency: coin,
				Free: a.Stocks,
				Frozen: a.FrozenStocks,
				Risk: a.Risk,
				Value: a.Value,
				Info: a.Info
			});
		}
		return re;
	}

	GetPosition(...args) {
		return this.rest.GetPosition(...args);
	}

	async GetNetPosition(...args) {
		let re = {
			Amount: 0
		};

		let positions = await this.GetPosition(...args);
		positions.map(p => {
			if (p.Type === 'Long') {
				re.Amount = N(re.Amount).add(p.Amount) * 1;
			} else {
				re.Amount = N(re.Amount).minus(p.Amount) * 1;
			}
		});
		return re;
	}

	GetTicker(...args) {
		return this.rest.GetTicker(...args);
	}

	GetOrder(...args) {
		return this.rest.GetOrder(...args);
	}

	GetOrders(...args) {
		return this.rest.GetOrders(...args);
	}

	Trade(direction, price, amount, Currency, BaseCurrency, ContractType, options) {
		if (['Long', 'Short', 'CloseLong', 'CloseShort'].indexOf(direction) === -1) {
			throw new Error('unknown direction: ' + direction + ', direction should be one of these values: Long Short CloseLong CloseShort');
		}
		ok( amount > 0, 'amount should greater than 0');

		if (typeof Currency === 'object') {
			options = Currency;
			Currency = undefined;
		}
		if (typeof BaseCurrency === 'object') {
			options = BaseCurrency;
			BaseCurrency = undefined;
		}
		if (typeof ContractType === 'object') {
			options = ContractType;
			ContractType = undefined;
		}
		if (!options) options = {};
		options.ClientId = this.GetClientId(options.ClientId);
		if (options.ClientId) options.ClientId = 'nw' + options.ClientId;
		
		console.log(this.GetName(), direction, price, amount, options);
		return this.rest.Trade(direction, price, amount, Currency, BaseCurrency, ContractType, options).then( r => {
			if (r && r.order_id) {
				return r.order_id;
			} else {
				console.log(r);
				throw new Error(r);
			}
		});
	}

	CancelOrder(...args) {
		return this.rest.CancelOrder(...args).then(result => {
			if (result && (result.result === true || result.result === 'true')) return true;
			if (result && result.error_code === 1050) return true;
			if (result && result.error_code === 1009) return true;
			throw new Error('cancel order failed, result: ' + JSON.stringify(result));
		});
	}

	CancelPendingOrders(...args) {
		console.log('cancelling pending orders...');
		return this.GetOrders(...args).then( orders => {
			console.log('cancelling', orders.length, 'orders');
			return Promise.all(orders.map( o => {
				return this.CancelOrder(o.Id, ...args);
			})).then( results => {
				console.log(results);
				return true;
			});
		});
	}

	GetDepth(...args) {
		return this.rest.GetDepth(...args);
	}

	GetMarkets() {
		return this.rest.GetMarkets();
	}

	GetTickers() {
		return this.rest.GetTickers();
	}

	GetRecords(minutes, Currency, BaseCurrency, ContractType) {
		return this.rest.GetRecords(minutes, Currency, BaseCurrency, ContractType);
	}

}


module.exports = OKEXV3_FUTURE;
