const fetch = require('node-fetch');
const N = require('precise-number');
const R = require('ramda');
const debug = require('debug')('exchange:okex:rest');
const agent = require('../../lib/agent');
const crypto = require('crypto');
const wait = require('delay');

class OKEX {

	constructor(options) {
		this.options = options;
		if (!this.options.rateLimiter) throw 'No rateLimiter in options';

		this.instrument_id_to_alias_cache = null;
		this.alias_to_instrument_id = {};

		this.aliasGot = false;
		this.gettingAlias = false;
	}

	async getAlias(instrument_id) {
		if (this.instrument_id_to_alias_cache === null) {
			let markets = await this.GetMarkets();
			this.instrument_id_to_alias_cache = {};
			markets.map(m => {
				this.instrument_id_to_alias_cache[m.Info.instrument_id] = m.ContractType;
				this.alias_to_instrument_id[`${m.Currency}-${m.BaseCurrency}-${m.ContractType}`] = m.Info.instrument_id;
			});
		}
		return this.instrument_id_to_alias_cache[instrument_id] || 'unknown';
	}

	async waitUntilAliasReady() {
		if (this.aliasGot) return true;
		if (!this.gettingAlias) {
			this.gettingAlias = true;
			await this.getAlias();
			this.aliasGot = true;
			this.gettingAlias = false;
		} else {
			let startTime = Date.now();
			while (this.gettingAlias) {
				if (Date.now() - startTime > 20000) throw new Error('waitUntilAliasReady timeout');
				await wait(50);
			}
		}
		return true;
	}

	async fetch(method, path, postBody = '', isPublic = false) {
		await this.options.rateLimiter.wait();
		if (method === 'POST') {
			if (typeof postBody === 'object') postBody = JSON.stringify(postBody);
		}

		debug('<<', method, 'https://www.okex.com' + path, 'BODY:', postBody);

		let timestamp = new Date().toISOString();
		let headers = {
			'Content-Type': 'application/json'
		};
		if (!isPublic) {
			headers['OK-ACCESS-KEY'] = this.options.Key;
			headers['OK-ACCESS-SIGN'] = this.sign(timestamp, method, path, postBody);
			headers['OK-ACCESS-TIMESTAMP'] = timestamp;
			headers['OK-ACCESS-PASSPHRASE'] = this.options.Pass;
		}

		let options = {
			method,
			timeout: method === 'GET' ? 4000 : 4000,
			headers,
			agent: agent.https
		};

		if (postBody) options.body = postBody;

		return fetch('https://www.okex.com' + path, options).then(res => res.text()).then(t => {
			debug('>> ' + t);

			if (!t) return Promise.reject('OKEX returns empty: ' + path);
			try {
				let d = JSON.parse(t);
				return Promise.resolve(d);
			} catch ( err ) {
				return Promise.reject('OKEX JSON parse error: ' + t);
			}
		}).then(json => {
			if (json && json.message === 'System error') {
				throw new Error('OKEX: System error');
			} else if (json && json.code && json.message) {
				let err = new Error(JSON.stringify(json));
				err.error_code = json.code;
				err.original = json;
				err.options = options;
				err.path = path;
				throw err;
			} else {
				return Promise.resolve(json);
			}
		}).catch(err => {
			if (err && err.type === 'request-timeout') {
				throw path + ' timeout';
			}
			throw err;
		});
	}

	sign(timestamp, method, path, body) {
		let presign = timestamp + method + path;
		if (body) presign += body;
		return crypto.createHmac('sha256', this.options.Secret).update(presign).digest('base64');
	}

	_getSymbol(Currency, BaseCurrency, ContractType) {
		if (this.options.ContractId) {
			if (!Currency) return this.options.ContractId;
			if (!BaseCurrency) BaseCurrency = this.options.BaseCurrency;
			let { Currency: c, BaseCurrency: bc } = this._parseInstrumentId(this.options.ContractId);
			if (Currency === c && BaseCurrency === bc) return this.options.ContractId;
		}
		if (!Currency) Currency = this.options.Currency;
		if (!BaseCurrency) BaseCurrency = this.options.BaseCurrency;
		if (!ContractType) ContractType = this.options.ContractType;
		let symbol = this.alias_to_instrument_id[`${Currency}-${BaseCurrency}-${ContractType}`];
		if (!symbol) throw new Error('can not build instrument_id from ' + Currency + ' ' + BaseCurrency + ' ' + ContractType);
		return symbol;
	}

	_parseSymbol(...args) {
		return this._parseInstrumentId(this._getSymbol(...args));
	}

	_parseInstrumentId(instrument_id) {
		if (this.instrument_id_to_alias_cache === null) throw new Error('need to call getAlias() first');
		let [Currency, BaseCurrency] = String(instrument_id).split('-');
		let ContractType = this.instrument_id_to_alias_cache[instrument_id];
		return { Currency, BaseCurrency, ContractType, ContractId: instrument_id};
	}

	async GetTicker(Currency, BaseCurrency, ContractType) {
		await this.waitUntilAliasReady();
		let instrument_id = this._getSymbol(Currency, BaseCurrency, ContractType);
		return this.fetch('GET', `/api/futures/v3/instruments/${instrument_id}/ticker`, null, true).then(data => {
			return Promise.resolve({
				High: N.parse(data.high_24h),
				Low: N.parse(data.low_24h),
				Buy: N.parse(data.best_bid),
				Sell: N.parse(data.best_ask),
				Last: N.parse(data.last),
				Volume: N.parse(data.volume_24h),
				Time: new Date(data.timestamp).getTime(),
				ContractId: data.instrument_id,
				...this._parseInstrumentId(data.instrument_id)
			});
		});
	}

	async GetPublicTrades(Currency, BaseCurrency, ContractType) {
		await this.waitUntilAliasReady();
		let instrument_id = this._getSymbol(Currency, BaseCurrency, ContractType);
		let data = await this.fetch('GET', `/api/futures/v3/instruments/${instrument_id}/trades`, null, true);
		let trades = [];
		if (data && data.length > 0) {
			data.map(o => {
				trades.push({
					Id: o.trade_id,
					Time: new Date(o.timestamp).getTime(),
					Price: N.parse(o.price),
					Amount: N.parse(o.qty),
					Type: o.side === 'buy' ? 'Long' : 'Short',
					...this._parseInstrumentId(instrument_id)
				});
			});
		};
		return trades;
	}

	async GetAccount(Currency) {
		if (!Currency) Currency = this.options.Currency;
		await this.waitUntilAliasReady();
		return this.fetch('GET', `/api/futures/v3/accounts/${Currency.toLowerCase()}`, null).then(data => {
			return this._parseAccount(data);
		});
	}

	_parseAccount(info) {
		let frozen = N(info.margin || 0).add(info.margin_frozen || 0) * 1;
		return {
			Balance: 0,
			FrozenBalance: 0,
			Stocks: N(info.equity).minus(frozen) * 1,
			FrozenStocks: frozen,
			Value: N.parse(info.equity),
			Risk: N.parse(info.margin_ratio || 0),
			Profit: N(info.realized_pnl || 0).add(info.unrealized_pnl || 0) * 1,
			Time: Date.now(),
			Info: info,
			...this._parseSymbol()
		};
	}

	async GetPosition(Currency, BaseCurrency, ContractType) {
		await this.waitUntilAliasReady();
		let d = await this.fetch('GET', `/api/futures/v3/${this._getSymbol(Currency, BaseCurrency, ContractType)}/position`);
		if (d && d.result && Array.isArray(d.holding)) {
			let re = [];
			d.holding.map(h => {
				if (h.long_qty && h.long_qty * 1 > 0) {
					re.push({
						MarginLevel: N.parse(h.leverage),
						Amount: N.parse(h.long_avail_qty),
						FrozenAmount: N(h.long_qty).minus(h.long_avail_qty) * 1,
						Price: N.parse(h.long_settlement_price),
						Profit: N.parse(h.long_unrealised_pnl),
						Type: 'Long',
						...this._parseInstrumentId(h.instrument_id),
						Info: h
					});
				}

				if (h.short_qty && h.short_qty * 1 > 0) {
					re.push({
						MarginLevel: N.parse(h.leverage),
						Amount: N.parse(h.short_avail_qty),
						FrozenAmount: N(h.short_qty).minus(h.short_avail_qty) * 1,
						Price: N.parse(h.short_settlement_price),
						Profit: N.parse(h.short_unrealised_pnl),
						Type: 'Short',
						...this._parseInstrumentId(h.instrument_id),
						Info: h
					});
				}
			});

			return re;
		}

		throw new Error('error position response: ' + JSON.stringify(d));
	}

	async GetOrder(orderId, Currency, BaseCurrency, ContractType) {
		await this.waitUntilAliasReady();
		return this.fetch('GET', `/api/futures/v3/orders/${this._getSymbol(Currency, BaseCurrency, ContractType)}/${orderId}`).then(o => {
			return this._transform_order(o);
		});
	}

	_transform_order(o) {
		return {
			Id: String(o.order_id),
			Price: N.parse(o.price),
			Amount: N.parse(o.size),
			DealAmount: N.parse(o.filled_qty),
			Status: this._order_status(o.status),
			Type: this._order_type(o.type),
			AvgPrice: N.parse(o.price_avg),
			Fee: N.parse(o.fee),
			Time: new Date(o.timestamp).getTime(),
			...this._parseInstrumentId(o.instrument_id),
			Info: o
		};
	}

	_order_status( status ) {
		//订单状态(0等待成交 1部分成交 2全部成交 -1撤单 4撤单处理中 5撤单中)
		let arr = {
			'-2': 'Cancelled',
			'-1': 'Cancelled',
			'0': 'Pending',
			'1': 'Partial',
			'2': 'Closed',
			'3': 'Pending'
		};
		return arr[`${status}`];
	}

	_order_type( type ) {
		//订单类型 1：开多 2：开空 3：平多 4： 平空
		let arr = {
			'1': 'Long',
			'2': 'Short',
			'3': 'CloseLong',
			'4': 'CloseShort'
		};
		return arr[`${type}`];
	}

	async GetOrders(Currency, BaseCurrency, ContractType) {
		await this.waitUntilAliasReady();
		let orders = await this.fetch('GET', `/api/futures/v3/orders/${this._getSymbol(Currency, BaseCurrency, ContractType)}?status=6`);
		if (orders.order_info) orders = orders.order_info;
		return orders.map(o => {
		 	return this._transform_order(o);
		});
	}

	async CancelOrder(orderId, Currency, BaseCurrency, ContractType) {
		await this.waitUntilAliasReady();
		return this.fetch('POST', `/api/futures/v3/cancel_order/${this._getSymbol(Currency, BaseCurrency, ContractType)}/${orderId}`);
	}

	async Trade(direction, price, amount, Currency, BaseCurrency, ContractType, options) {
		await this.waitUntilAliasReady();
		let instrument_id = this._getSymbol(Currency, BaseCurrency, ContractType);
		let params = {
			instrument_id,
			type: (['Long', 'Short', 'CloseLong', 'CloseShort'].indexOf(direction) + 1) + '',
			order_type: this.getOrderType(options.Type),
			price,
			size: amount,
			match_price: price === -1 ? '1' : '0',
			leverage: this.options.MarginLevel
		};
		if (price === -1) {
			delete(params.price);
		}

		if (options.ClientId) {
			params.client_oid = options.ClientId;
		}

		return this.fetch('POST', `/api/futures/v3/order`, params);
	}

	getOrderType(Type) {
		const types = {
			GTC: 0,
			PO: 1,
			FOK: 2,
			IOC: 3
		};
		if (!Type) return this.options.MakerMode ? types.PO : types.GTC;
		return types[Type] || types.GTC;
	}

	async GetDepth(Currency, BaseCurrency, ContractType, size) {
		await this.waitUntilAliasReady();
		if (!size) size = 30;
		let instrument_id = this._getSymbol(Currency, BaseCurrency, ContractType);
		return this.fetch('GET', `/api/futures/v3/instruments/${instrument_id}/book?size=${size}`, null, true).then(data => {
			if (!data || !data.bids || !data.asks) throw new Error('get okex depth error ' + JSON.stringify(data));

			let asks = [];
			let bids = [];

			for (let i = 0;i < data.bids.length;i++) {
				bids.push({
					Price: data.bids[i][0] * 1,
					Amount: data.bids[i][1] * 1
				});
			}

			for (let i = 0;i < data.asks.length;i++) {
				asks.push({
					Price: data.asks[i][0] * 1,
					Amount: data.asks[i][1] * 1
				});
			}

			bids.sort((a, b) => {
				let diff = a.Price - b.Price;
				return diff > 0 ? -1 : diff === 0 ? 0 : 1;
			});

			asks.sort((a, b) => {
				let diff = a.Price - b.Price;
				return diff > 0 ? 1 : diff === 0 ? 0 : -1;
			});

			return Promise.resolve({
				Asks: R.sort( R.ascend( R.prop('Price') ), asks),
				Bids: R.sort( R.descend( R.prop('Price') ), bids),
				Time: new Date(data.timestamp).getTime(),
				...this._parseInstrumentId(instrument_id)
			});
		});
	}

	async GetTickers() {
		await this.waitUntilAliasReady();
		return this.fetch('GET', `/api/futures/v3/instruments/ticker`, null, true).then(data => {
			return data.map(data => {
				return {
					High: N.parse(data.high_24h),
					Low: N.parse(data.low_24h),
					Buy: N.parse(data.best_bid),
					Sell: N.parse(data.best_ask),
					Last: N.parse(data.last),
					Volume: N.parse(data.volume_24h),
					Time: new Date(data.timestamp).getTime(),
					ContractId: data.instrument_id,
					...this._parseInstrumentId(data.instrument_id)
				};
			});
		});
	}

	async GetMarkets() {
		let arr = await this.fetch('GET', `/api/futures/v3/instruments`, null, true);
		return arr.map(a => {
			let d = new Date(a.delivery + 'T08:00:00.000Z');
			return {
				Currency: a.underlying_index,
				BaseCurrency: a.quote_currency,
				ContractType: a.alias,
				ContractId: a.instrument_id,
				DeliveryTime: d.getTime(),
				UnitValue: N.parse(a.contract_val),
				UnitCurrency: a.contract_val_currency,
				Decimals: Math.log10(N(1).div(a.tick_size)),
				StockDecimals: Math.log10(N(1).div(a.trade_increment)),
				MinTradeAmount: 1,
				Info: a
			};
		});
	}

	async GetRecords(minutes = 1, Currency, BaseCurrency, ContractType) {
		await this.waitUntilAliasReady();
		let arr = await this.fetch('GET', `/api/futures/v3/instruments/${this._getSymbol(Currency, BaseCurrency, ContractType)}/candles?granularity=${minutes * 60}`, null, true);
		return arr.map(k => {
			let [timestamp, open, high, low, close, volume, currency_volume] = k;
			return {
				Open: N.parse(open),
				High: N.parse(high),
				Low: N.parse(low),
				Close: N.parse(close),
				Volume: N.parse(volume),
				CurrencyVolume: N.parse(currency_volume),
				Time: new Date(timestamp).getTime()
			};
		}).sort((a, b) => {
			return a.Time - b.Time;
		}).slice(-200);
	}
}

module.exports = OKEX;
