const N = require('precise-number');
const WebSocket = require('../../lib/auto-reconnect-ws.js');
const OrderBook = require('../../lib/orderbook.js');
const crypto = require('crypto');
const Events = require('events');
const tc = require('turbocolor');
const pako = require('pako');
const wait = require('delay');

class OKEXV3_FUTURE_WS extends Events {
	constructor(options, rest) {
		super();
		this.options = options;
		this.rest = rest;
		this.lastPingTime = Date.now();
		this.wsReady = false;
		this.orderbooks = {};
		this.msgTime = 0;
		this.DepthMode = this.options.WSDepthMode || 'depth5';
		this.DepthSize = this.options.WSDepthSize || 10;
		this.DepthCacheRange = this.options.WSDepthCacheRange || 10000;
		this.boot().catch(err => {
			console.error('okex future websocket connect error', err);
			process.exit();
		});
	}

	async boot() {
		await this.rest.waitUntilAliasReady();

		this.symbol = this.rest._getSymbol();
		this.ws = new WebSocket('wss://real.okex.com:8443/ws/v3');

		//remember subscription commands
		this.subscriptionCommands = [];

		if (this.options.onDepth) {
			this.subscriptionCommands.push(JSON.stringify({
				op: 'subscribe',
				args: [`futures/${this.DepthMode}:${this.symbol}`]
			}));
		}

		if (this.options.onTicker) {
			this.subscriptionCommands.push(JSON.stringify({
				op: 'subscribe',
				args: [`futures/ticker:${this.symbol}`]
			}));
		}

		if (this.options.onPublicTrades) {
			this.subscriptionCommands.push(JSON.stringify({
				op: 'subscribe',
				args: [`futures/trade:${this.symbol}`]
			}));
		}

		this.ws.on('open', () => {
			this.wsReady = true;
			this.subscriptionCommands.map(cmd => {
				this.ws.send(cmd);
			});
			this.emit('connect');
		});

		this.ws.on('message', (data) => {
			this.wsReady = true;
			this.msgTime = Date.now();
			this.onMessage(data);
		});

		this.ws.on('pong', t => {
			this.wsReady = true;
			if (this.options.onPong) {
				this.options.onPong(t);
			}
		});

		this.ws.on('close', () => {
			this.emit('close');
			this.wsReady = false;
		});

		this.ws.on('error', () => {
			this.emit('close');
			this.wsReady = false;
		});

		setInterval(() => {
			if (this.wsReady) {
				this.lastPingTime = Date.now();
				this.ws.send('ping');
			}
		}, 5000);
	}

	async waitUntilWSReady() {
		if (!this.options.isWS) return true;
		let startTime = Date.now();
		while (!this.wsReady) {
			await wait(200);
			if (Date.now() - startTime > this.options.WSTimeout) {
				throw new Error('okex future websocket connection timeout');
			}
		}
		return true;
	}

	sign(timestamp, method, path, body) {
		let presign = timestamp + method + path;
		if (body) presign += body;
		return crypto.createHmac('sha256', this.options.Secret).update(presign).digest('base64');
	}

	onMessage(raw)  {
		if (typeof raw !== 'string') {
			raw = pako.inflateRaw(raw, {to: 'string'});
		}
		if (raw === 'pong') {
			this.onPong();
			return;
		}

		let message = null;
		try {
			message = JSON.parse(raw);
		} catch (err) {
			console.error('okex websocket message error: ' + raw);
			return;
		}
		if (!message) return;

		if (message.event === 'login') {
			if (message.success) {
				console.log(tc.green('okex websocket login success'));
				this.wsReady = true;
				this.emit('connect');
			} else {
				console.error(tc.red('okex websocket login failed'), message);
				process.exit();
				return;
			}
			return;
		}

		if (message.event === 'error') {
			console.error(message);
			return;
		}

		let { table, data } = message;
		if (!table || !data) return;

		if (table === 'futures/ticker') {
			this.onTicker(message);
		} else if (table === 'futures/depth5') {
			this.onDepth(message);
		} else if (table === 'futures/trade') {
			this.onPublicTrades(message);
		} else if (table === 'futures/depth_l2_tbt' || table === 'futures/depth') {
			this.onL2Depth(message);
		} else {
			console.log('unknown table:', table);
		}
	}

	onPong() {
		// let gap = Date.now() - this.lastPingTime;
		// console.log('okex websocket ping pong time:', gap + 'ms');
		this.lastPong = Date.now();
	}

	addSubscription(Currency, BaseCurrency, ContractType, type) {
		this.waitUntilWSReady().then(() => {
			let channel = '';
			if (type === 'Ticker') {
				channel = `futures/ticker:${this.rest._getSymbol(Currency, BaseCurrency, ContractType)}`;
			} else if (type === 'Depth') {
				channel = `futures/${this.DepthMode}:${this.rest._getSymbol(Currency, BaseCurrency, ContractType)}`;
			} else if (type === 'PublicTrades') {
				channel = `futures/trade:${this.rest._getSymbol(Currency, BaseCurrency, ContractType)}`;
			}

			if (!channel) return;

			let cmd = JSON.stringify({
				op: 'subscribe',
				args: [channel]
			});

			this.subscriptionCommands.push(cmd);
			this.ws.send(cmd);
		}).catch(err => {
			console.error('addSubscription got error', err);
		});
	}

	onDepth(data) {
		data = data.data[0];
		let {timestamp, asks, bids, instrument_id} = data;
		if (!asks || !bids || !instrument_id || !timestamp) return;

		asks = asks.map(pair => {
			return {
				Price: pair[0] * 1,
				Amount: pair[1] * 1
			};
		});

		bids = bids.map(pair => {
			return {
				Price: pair[0] * 1,
				Amount: pair[1] * 1
			};
		});

		let depth = {
			Time: new Date(timestamp).getTime(),
			Asks: asks,
			Bids: bids,
			...this.rest._parseInstrumentId(instrument_id)
		};

		this.options.onDepth(depth, this.msgTime);
	}

	onL2Depth(message) {
		let { data } = message;
		this.updateOrderBook(data);
	}

	updateOrderBook(data) {
		if (!data || data.length < 1) return;
		let { instrument_id, timestamp, asks, bids } = data[0];
		if (!this.orderbooks[instrument_id]) this.orderbooks[instrument_id] = new OrderBook();
		let orderbook = this.orderbooks[instrument_id];

		asks = asks.map(t => {
			let Price = N.parse(t[0]);
			let Amount = N.parse(t[1]);
			return {Price, Amount};
		});

		bids = bids.map(t => {
			let Price = N.parse(t[0]);
			let Amount = N.parse(t[1]);
			return {Price, Amount};
		});

		orderbook.update(asks, bids, new Date(timestamp).getTime());

		if (this.options.onDepth) {
			let depth = orderbook.get();
			this.options.onDepth({
				Asks: depth.Asks.slice(0, this.DepthSize),
				Bids: depth.Bids.slice(0, this.DepthSize),
				Time: depth.Time,
				...this.rest._parseInstrumentId(instrument_id)
			}, this.msgTime);
		}
	}

	onTicker({data}) {
		if (!data || data.length <= 0) return;
		let tick = data[0];
		let re = {
			High: N.parse(tick.high_24h),
			Low: N.parse(tick.low_24h),
			Close: N.parse(tick.last),
			Last: N.parse(tick.last),
			Volume: N.parse(tick.volume_24h),
			Time: new Date(tick.timestamp).getTime(),
			...this.rest._parseInstrumentId(tick.instrument_id)
		};
		if (this.options.onTicker) {
			this.options.onTicker(re);
		}
	}

	onPublicTrades({data}) {
		if (!data || data.length <= 0) return;
		let trades = data.map(t => {
			return {
				Amount: N.parse(t.qty),
				Price: N.parse(t.price),
				Type: t.side === 'buy' ? 'Buy' : 'Sell',
				Time: new Date(t.timestamp).getTime(),
				Id: t.trade_id,
				...this.rest._parseInstrumentId(t.instrument_id)
			};
		});
		this.options.onPublicTrades(trades);
	}
}

module.exports = OKEXV3_FUTURE_WS;
