const fetch = require('node-fetch');
const N = require('precise-number');
const R = require('ramda');
const debug = require('debug')('exchange:okex:rest');
const agent = require('../../lib/agent');
const crypto = require('crypto');

class OKEX {

	constructor(options) {
		this.options = options;
		this.symbol = this._getSymbol(options.Currency, options.BaseCurrency);
		if (!this.options.rateLimiter) throw 'No rateLimiter in options';
	}

	async fetch(method, path, postBody = '', isPublic = false) {
		await this.options.rateLimiter.wait();
		if (method === 'POST') {
			if (typeof postBody === 'object') postBody = JSON.stringify(postBody);
		}

		debug('<<', method, 'https://www.okex.com' + path, 'BODY:', postBody);

		let timestamp = new Date().toISOString();
		let headers = {
			'Content-Type': 'application/json'
		};
		if (!isPublic) {
			headers['OK-ACCESS-KEY'] = this.options.Key;
			headers['OK-ACCESS-SIGN'] = this.sign(timestamp, method, path, postBody);
			headers['OK-ACCESS-TIMESTAMP'] = timestamp;
			headers['OK-ACCESS-PASSPHRASE'] = this.options.Pass;
		}

		let options = {
			method,
			timeout: method === 'GET' ? 4000 : 4000,
			headers,
			agent: agent.https
		};

		if (postBody) options.body = postBody;

		return fetch('https://www.okex.com' + path, options).then(res => res.text()).then(t => {
			debug('>> ' + t);

			if (!t) return Promise.reject('OKEX returns empty: ' + path);
			try {
				let d = JSON.parse(t);
				return Promise.resolve(d);
			} catch ( err ) {
				return Promise.reject('OKEX JSON parse error: ' + t);
			}
		}).then(json => {
			if (json && json.message === 'System error') {
				throw new Error('OKEX: System error');
			} else if (json && json.code && json.message) {
				let err = new Error(JSON.stringify(json));
				err.error_code = json.code;
				err.original = json;
				err.options = options;
				err.path = path;
				throw err;
			} else {
				return Promise.resolve(json);
			}
		}).catch(err => {
			if (err && err.type === 'request-timeout') {
				throw path + ' timeout';
			}
			throw err;
		});
	}

	sign(timestamp, method, path, body) {
		let presign = timestamp + method + path;
		if (body) presign += body;
		return crypto.createHmac('sha256', this.options.Secret).update(presign).digest('base64');
	}

	_getSymbol(Currency, BaseCurrency) {
		if (!Currency) Currency = this.options.Currency;
		if (!BaseCurrency) BaseCurrency = this.options.BaseCurrency;
		return `${Currency}-${BaseCurrency}-SWAP`.toUpperCase();
	}

	_parseSymbol(Currency, BaseCurrency) {
		if (!Currency) Currency = this.options.Currency;
		if (!BaseCurrency) BaseCurrency = this.options.BaseCurrency;
		return { Currency, BaseCurrency, ContractType: 'SWAP' };
	}

	_parseSymbolString(symbol) {
		if (!symbol) symbol = this.symbol;
		let [Currency, BaseCurrency, ContractType] = String(symbol).toUpperCase().split('-');
		if (!Currency) Currency = this.options.Currency;
		if (!BaseCurrency) BaseCurrency = this.options.BaseCurrency;
		if (!ContractType) ContractType = 'SWAP';
		return { Currency, BaseCurrency, ContractType };
	}

	GetTicker(Currency, BaseCurrency) {
		return this.fetch('GET', `/api/swap/v3/instruments/${this._getSymbol(Currency, BaseCurrency)}/ticker`, null, true).then(data => {
			return Promise.resolve({
				High: N.parse(data.high_24h),
				Low: N.parse(data.low_24h),
				Buy: N.parse(data.best_bid),
				Sell: N.parse(data.best_ask),
				Last: N.parse(data.last),
				Volume: N.parse(data.volume_24h),
				Time: new Date(data.timestamp).getTime(),
				ContractId: data.instrument_id,
				...this._parseSymbol(Currency, BaseCurrency)
			});
		});
	}

	async GetPublicTrades(Currency, BaseCurrency) {
		let data = await this.fetch('GET', `/api/swap/v3/instruments/${this._getSymbol(Currency, BaseCurrency)}/trades`, null, true);
		let trades = [];
		if (data && data.length > 0) {
			data.map(o => {
				trades.push({
					Id: o.trade_id,
					Time: new Date(o.timestamp).getTime(),
					Price: N.parse(o.price),
					Amount: N.parse(o.size),
					Type: o.side === 'buy' ? 'Long' : 'Short',
					...this._parseSymbol(Currency, BaseCurrency)
				});
			});
		};
		return trades;
	}

	GetAccount(Currency, BaseCurrency) {
		return this.fetch('GET', `/api/swap/v3/${this._getSymbol(Currency, BaseCurrency)}/accounts`, null).then(data => this._parseAccount(data.info));
	}

	_parseAccount(info) {
		let frozen = N(info.margin || 0).add(info.margin_frozen || 0) * 1;
		return {
			Balance: 0,
			FrozenBalance: 0,
			Stocks: N(info.equity).minus(frozen) * 1,
			FrozenStocks: frozen,
			Value: N.parse(info.equity),
			Risk: N.parse(info.margin_ratio || 0),
			Profit: N(info.realized_pnl || 0).add(info.unrealized_pnl || 0) * 1,
			Time: new Date(info.timestamp).getTime(),
			...this._parseSymbolString(info.instrument_id),
			Info: info
		};
	}

	async GetPosition(Currency, BaseCurrency) {
		let d = await this.fetch('GET', `/api/swap/v3/${this._getSymbol(Currency, BaseCurrency)}/position`);
		if (d && Array.isArray(d.holding)) {
			let re = [];
			d.holding.map(h => {
				re.push({
					MarginLevel: N.parse(h.leverage),
					Amount: N.parse(h.position),
					FrozenAmount: N(h.position).minus(h.avail_position) * 1,
					Price: N.parse(h.avg_cost),
					Profit: N.parse(h.realized_pnl),
					Type: h.side === 'short' ? 'Short' : 'Long',
					Time: new Date(h.timestamp).getTime(),
					...this._parseSymbolString(h.instrument_id),
					Info: h
				});
			});
			return re;
		}

		throw new Error('error position response: ' + JSON.stringify(d));
	}

	GetOrder(orderId, Currency, BaseCurrency) {
		return this.fetch('GET', `/api/swap/v3/orders/${this._getSymbol(Currency, BaseCurrency)}/${orderId}`);
	}

	async GetOrders(Currency, BaseCurrency) {
		let results = this.fetch('GET', `/api/swap/v3/orders/${this._getSymbol(Currency, BaseCurrency)}?status=6`);
		return results.order_info;
	}

	CancelOrder(orderId, Currency, BaseCurrency) {
		return this.fetch('POST', `/api/swap/v3/cancel_order/${this._getSymbol(Currency, BaseCurrency)}/${orderId}`);
	}

	async Trade(direction, price, amount, Currency, BaseCurrency, ContractType, options) {
		let params = {
			instrument_id: this._getSymbol(Currency, BaseCurrency),
			type: (['Long', 'Short', 'CloseLong', 'CloseShort'].indexOf(direction) + 1) + '',
			order_type: this.getOrderType(options.Type),
			price,
			size: amount,
			match_price: price === -1 ? '1' : '0',
			lever_rate: this.options.MarginLevel
		};
		if (price === -1) {
			delete(params.price);
		}

		if (options.ClientId) {
			params.client_oid = options.ClientId;
		}

		return this.fetch('POST', `/api/swap/v3/order`, params);
	}

	getOrderType(Type) {
		const types = {
			GTC: 0,
			PO: 1,
			FOK: 2,
			IOC: 3
		};
		if (!Type) return this.options.MakerMode ? types.PO : types.GTC;
		return types[Type] || types.GTC;
	}

	GetDepth(Currency, BaseCurrency, size) {
		if (!size) size = 30;
		return this.fetch('GET', `/api/swap/v3/instruments/${this._getSymbol(Currency, BaseCurrency)}/depth?size=${size}`, null, true).then(data => {

			if (!data || !data.bids || !data.asks) throw new Error('get okex ' + this.symbol + ' depth error ' + JSON.stringify(data));

			let asks = [];
			let bids = [];

			for (let i = 0;i < data.bids.length;i++) {
				bids.push({
					Price: data.bids[i][0] * 1,
					Amount: data.bids[i][1] * 1
				});
			}

			for (let i = 0;i < data.asks.length;i++) {
				asks.push({
					Price: data.asks[i][0] * 1,
					Amount: data.asks[i][1] * 1
				});
			}

			bids.sort((a, b) => {
				let diff = a.Price - b.Price;
				return diff > 0 ? -1 : diff === 0 ? 0 : 1;
			});

			asks.sort((a, b) => {
				let diff = a.Price - b.Price;
				return diff > 0 ? 1 : diff === 0 ? 0 : -1;
			});

			return Promise.resolve({
				Asks: R.sort( R.ascend( R.prop('Price') ), asks),
				Bids: R.sort( R.descend( R.prop('Price') ), bids),
				Time: new Date(data.time).getTime(),
				...this._parseSymbol(Currency, BaseCurrency)
			});
		});
	}

	async GetTickers() {
		return [ await this.GetTicker() ];
	}

	async GetMarkets() {
		let arr = await this.fetch('GET', `/api/swap/v3/instruments`, null, true);
		return arr.map(a => {
			return {
				Currency: a.coin,
				BaseCurrency: a.quote_currency,
				ContractType: 'SWAP',
				ContractId: a.instrument_id,
				UnitValue: Math.round(a.contract_val),
				UnitCurrency: 'USD',
				Decimals: Math.log10(N(1).div(a.tick_size)),
				StockDecimals: Math.log10(N(1).div(a.size_increment)),
				MinTradeAmount: 1,
				Info: a
			};
		});
	}

	async GetRecords(minutes = 1, Currency, BaseCurrency) {
		let arr = await this.fetch('GET', `/api/swap/v3/instruments/${this._getSymbol(Currency, BaseCurrency)}/candles?granularity=${minutes * 60}`, null, true);
		return arr.map(k => {
			let [timestamp, open, high, low, close, volume, currency_volume] = k;
			return {
				Open: N.parse(open),
				High: N.parse(high),
				Low: N.parse(low),
				Close: N.parse(close),
				Volume: N.parse(volume),
				CurrencyVolume: N.parse(currency_volume),
				Time: new Date(timestamp).getTime()
			};
		}).sort((a, b) => {
			return a.Time - b.Time;
		}).slice(-200);
	}
}

module.exports = OKEX;
