const N = require('precise-number');

let utils = {

	getPriceFromDepthByAmount(entries, amount) {
		return utils._getPriceFromDepth(entries, 'amount', amount);
	},

	getPriceFromDepthByMoney(entries, money) {
		return utils._getPriceFromDepth(entries, 'money', money);
	},

	_getPriceFromDepth(entries, type, tolerance) {
		if (!entries || entries.length === 0) throw 'entries is empty';
		let accumulatedAmount = 0;
		for (let i = 0; i < entries.length; i++) {
			let entry = entries[i];

			if (type === 'amount') {
				accumulatedAmount += entry.Amount;
			} else {
				accumulatedAmount += entry.Amount * entry.Price;
			}
			if ( tolerance <= accumulatedAmount) {
				return entry.Price;
			}
		}
		return entries[entries.length - 1].Price;
	},

	_isBidsOrAsks(entries) {
		if (!entries || entries.length <= 1) throw new Error('no entries');
		return entries[0].Price > entries[entries.length - 1].Price ? 'Bids' : 'Asks';
	},

	getTakerInfoFromDepthByPrice(entries, price) {
		let type = utils._isBidsOrAsks(entries);
		let re = entries.reduce((info, e) => {
			if ((type === 'Bids' && e.Price >= price) || (type === 'Asks' && e.Price <= price)) {
				info.Amount = N(info.Amount).add(e.Amount) * 1;
				info.Money = N(info.Money).add(N(e.Amount).multiply(e.Price)) * 1;
				info.Price = e.Price;
			}
			return info;
		}, { Money:0, Amount:0, Price:0 });
		re.AvgPrice = (re.Amount > 0) ? N.div(re.Money, re.Amount) : 0;
		return re;
	},

	getTakerInfoFromDepthByAmount(entries, amount) {
		let re = entries.reduce((info, e) => {
			if (info.Amount < amount) {
				info.Amount = N(info.Amount).add(e.Amount) * 1;
				info.Money = N(info.Money).add(N(e.Amount).multiply(e.Price)) * 1;
				info.Price = e.Price;
			}
			return info;
		}, { Money:0, Amount:0, Price:0 });
		re.AvgPrice = (re.Amount > 0) ? N.div(re.Money, re.Amount) : 0;
		return re;
	},

	combineDepth(depth, combineRatio) {
		if (!depth || !depth.Asks || !depth.Bids) return { Asks: [], Bids: [] };
		let midPrice = (depth.Asks[0].Price + depth.Bids[0].Price) / 2;
		let gap = utils.findBestGap(midPrice * combineRatio);
		return {
			Asks: utils.combineEntries(depth.Asks, gap),
			Bids: utils.combineEntries(depth.Bids, gap)
		};
	},

	combineEntries(entries, gap) {
		if (!entries || entries.length === 0) return [];
		let isAsks = entries[0].Price < entries[1].Price;
		let roundFunc = isAsks ? Math.ceil : Math.floor;
		let multiplier = gap.e >= 0 ? Math.pow(10, gap.e) : 1 / Math.pow(10, -gap.e);
		const gapPrice = multiplier * gap.m;
		multiplier = 1 / gapPrice;
		let lastPrice = roundFunc(entries[0].Price * multiplier);
		let lastAmount = 0;
		let re = [];
		for (let i = 0; i < entries.length - 1; i++) {
			let nextPrice = roundFunc(entries[i + 1].Price * multiplier);
			let diff = Math.abs(nextPrice - lastPrice);
			lastAmount += entries[i].Amount;
			if (diff > 0) {
				lastPrice = nextPrice;
				re.push({
					Price: roundFunc(entries[i].Price * multiplier),
					Amount: lastAmount
				});
				lastAmount = 0;
			}
		}

		let price = roundFunc(entries[entries.length - 1].Price * multiplier);
		if (lastPrice === entries[entries.length - 1].Price) {
			price += gapPrice * multiplier * (isAsks ? 1 : -1);
		}
		re.push({
			Price: price,
			Amount: lastAmount + entries[entries.length - 1].Amount
		});
		re = re.map(r => {
			r.Price = r.Price / multiplier;
			return r;
		});
		return re;
	},

	findBestGap(gap) {
		for (let i = 10; i >= -10; i--) {
			const m = i >= 0 ? Math.pow(10, i) : 1 / Math.pow(10, -i);
			if (gap >= m && gap < m * 10) {
				if (gap >= m * 5) {
					return { e: i, m: 5};
				} else if (gap >= m * 2) {
					return { e: i, m: 2};
				} else {
					return { e: i, m: 1};
				}
			}
		}
	}
};

module.exports = utils;